#!/usr/bin/env python
#
#   (C) 2001 by Argonne National Laboratory.
#       See COPYRIGHT in top-level directory.
#

import os, sys, socket, commands, time
sys.path.append("..")         # where to find mpdlib, etc.
from mpdlib import MPDTest    # do this AFTER extending sys.path

clusterHosts = []
# clusterHosts = ['bp400','bp401','bp402','bp403']
MPI_srcdir = "/home/rbutler/mpich"
MPI_1_dir  = "/home/rbutler/mpich1i"
USER_INSTALL_DIR = "/tmp/usermpd"
ROOT_INSTALL_DIR = "/tmp/rootmpd"
os.environ['PATH'] = "..:" + os.environ['PATH']
PYEXT = ".py"
HFILE = "temph"
os.environ['MPD_CON_EXT'] = "testing"


testsToRun = [ 0 for x in range(0,10000) ]         # init all to 0; override below

# for i in range(0,10000):    testsToRun[i] = 1    # run all tests
for i in range(0,1000):     testsToRun[i] = 1    # run tests for mpd
# for i in range(1000,2000):  testsToRun[i] = 1    # run tests for mpiexec
# for i in range(2000,3000):  testsToRun[i] = 1    # run MPI tests
# for i in range(3000,4000):  testsToRun[i] = 1    # run console pgm tests
# for i in range(8000,9000):  testsToRun[i] = 1    # run misc tests
# for i in range(9000,10000): testsToRun[i] = 1    # run root tests
# testsToRun[3003] = 1                               # run this one test


if 1 in testsToRun[0:1000]:
    print "mpd tests---------------------------------------------------"

if testsToRun[0]:
    # test: simple with 1 mpd (mpdboot uses mpd's -e and -d options)
    print "TEST -e -d"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %s" % (PYEXT,NMPDS) )
    expout = 'hello\nhello\nhello\n'
    mpdtest.run(cmd="mpiexec%s -n 3 /bin/echo hello" % (PYEXT), chkOut=1, expOut=expout )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1]:
    # test: simple with 2 mpds on same machine (mpdboot uses mpd's -n option)
    print "TEST -n"
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        temph = open(HFILE,'w')
        for i in range(NMPDS): print >>temph, '%s' % (socket.gethostname())
        temph.close()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.system("mpdboot%s -1 -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = 'hello\nhello\n'
        mpdtest.run(cmd="mpiexec%s -n 2 /bin/echo hello" % (PYEXT), chkOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[2]:
    # test: simple with 3 mpds on 3 machines
    print "TEST simple hello msg on 3 nodes"
    NMPDS = 3
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.system("mpdboot%s -1 -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = 'hello\nhello\nhello\n'
        mpdtest.run(cmd="mpiexec%s -n 3 /bin/echo hello" % (PYEXT), chkOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[3]:
    # test: simple 2 mpds on local machine (-l, -h, and -p option)
    print "TEST -l, -h, and -p"
    NMPDS = 3
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.system("mpd%s -d -l 12345" % (PYEXT) )
        os.system("mpd%s -d -n -h %s -p 12345" % (PYEXT,socket.gethostname()) )
        expout = 'hello\nhello\nhello\n'
        mpdtest.run(cmd="mpiexec%s -n 3 /bin/echo hello" % (PYEXT), chkOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[4]:
    # test: simple with 2 mpds on 2 machines  (--ncpus option)
    print "TEST --ncpus"
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, "%s:2" % (host)
        temph.close()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.system("mpdboot%s -f %s -n %d --ncpus=2" % (PYEXT,HFILE,NMPDS) )
        myHost = socket.gethostname()
        expout = '0: %s\n1: %s\n2: %s\n3: %s\n' % (myHost,myHost,clusterHosts[0],clusterHosts[0])
        mpdtest.run(cmd="mpiexec%s -l -n 4 /bin/hostname" % (PYEXT), chkOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[5]:
    # test: simple with 2 mpds on 2 machines  (--ifhn option)
    #   this is not a great test, but shows working with real ifhn, then failure with 127.0.0.1
    print "TEST minimal use of --ifhn"
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        temph = open(HFILE,'w')
        for host in clusterHosts:
            hostinfo = socket.gethostbyname_ex(host)
            IP = hostinfo[2][0]
            print >>temph, '%s ifhn=%s' % (host,IP)
        temph.close()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        hostinfo = socket.gethostbyname_ex(socket.gethostname())
        IP = hostinfo[2][0]
        os.system("mpdboot%s -f %s -n %d --ifhn=%s" % (PYEXT,HFILE,NMPDS,IP) )
        expout = 'hello\nhello\n'
        mpdtest.run(cmd="mpiexec%s -n 2 /bin/echo hello" % (PYEXT), chkOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        ## redo the above test with a local ifhn that should cause failure
        lines = commands.getoutput("mpdboot%s -f %s -n %d --ifhn=127.0.0.1" % (PYEXT,HFILE,NMPDS) )
        if len(lines) > 0:
            if lines.find('failed to handshake') < 0:
                print "probable error in ifhn test using 127.0.0.1; printing lines of output next:"
                print lines
                sys.exit(-1)
        os.unlink(HFILE)

if testsToRun[6]:
    print "TEST MPD_CON_INET_HOST_PORT"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    import random
    port = random.randrange(2000,9000)
    print "    using port", port
    os.environ['MPD_CON_INET_HOST_PORT'] = "localhost:%d" % (port)
    os.system("mpd.py &")
    time.sleep(1)  ##     time to get going
    expout = ['0: hello']
    rv = mpdtest.run(cmd="mpiexec%s -l -n 1 echo hello" % (PYEXT), expOut=expout,grepOut=1)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.unsetenv('MPD_CON_INET_HOST_PORT')   # get rid of this for later tests


if 1 in testsToRun[1000:2000]:
    print "mpiexec tests-------------------------------------------"

if testsToRun[1000]:
    print "TEST -machinefile"
    NMPDS = 4
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        tempm = open('tempm','w')
        for host in clusterHosts: print >>tempm, '%s:2 ifhn=%s' % (host,host)
        tempm.close()
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.system("mpdboot%s -1 -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = '0: bp400\n1: bp400\n2: bp401\n3: bp401\n'   # 2 per host because of :2's in tempm
        mpdtest.run(cmd="mpiexec%s -l -machinefile %s -n 4 hostname" % (PYEXT,'tempm'), chkOut=1, expOut=expout)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.system("rm -f tempm %s" % (HFILE))

if testsToRun[1001]:
    print "TEST -file"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    tempfilename = '/tmp/%s_tempin2' % (os.environ['USER'])
    f = open(tempfilename,'w')
    print >>f, """
       <create-process-group totalprocs='3'>
          <process-spec
	       range='0-2'
	       exec='/bin/echo'>
            <arg idx='1' value="hello"> </arg>
            <arg idx='2' value="again"> </arg>
          </process-spec>
       </create-process-group>
    """
    f.close()
    expout = 'hello again\nhello again\nhello again\n'
    mpdtest.run(cmd="mpiexec%s -file %s" % (PYEXT,tempfilename),chkOut=1,expOut=expout)
    os.unlink(tempfilename)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    
if testsToRun[1002]:
    print "TEST -configfile"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    tempfilename = '/tmp/%s_tempin2' % (os.environ['USER'])
    f = open(tempfilename,'w')
    print >>f, "-l\n-n 1 echo hello there\n-n 1 echo hello again"
    f.close()
    expout = '0: hello there\n1: hello again\n'
    mpdtest.run(cmd="mpiexec%s -configfile %s" % (PYEXT,tempfilename),chkOut=1,expOut=expout)
    os.unlink(tempfilename)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    
if testsToRun[1003]:
    print "TEST -l"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = '0: hello\n1: bye\n'
    mpdtest.run(cmd="mpiexec%s -l -n 1 echo hello : -n 1 echo bye" % (PYEXT),chkOut=1,expOut=expout)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1004]:
    print "TEST -m"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = '0-1:  hello\n'
    mpdtest.run(cmd="mpiexec%s -m -n 2 echo hello" % (PYEXT),chkOut=1,expOut=expout)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1005]:
    print "TEST -ecfn"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = 'hello\n'
    rv = mpdtest.run(cmd="mpiexec%s -ecfn tempxout -n 1 echo hello" % (PYEXT),chkOut=1,expOut=expout)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    linesAsStr = commands.getoutput("cat tempxout")
    os.unlink("tempxout")
    if linesAsStr.find('exit-codes') < 0:
        print "ecfn: Failed to create correct contents of xml file:"
        print linesAsStr
        sys.exit(-1)

if testsToRun[1006]:
    print "TEST -s"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n 1" % (PYEXT) )
    expin = 'hello\n'
    expout = '0: hello\n'
    rv = mpdtest.run(cmd="mpiexec%s -l -n 4 cat -" % (PYEXT),
                     expOut=expout,chkOut=1,expIn=expin)
    expout = '0: hello\n1: hello\n2: hello\n3: hello\n'
    rv = mpdtest.run(cmd="mpiexec%s -l -s 0-3 -n 4 cat -" % (PYEXT),
                     expOut=expout,chkOut=1,expIn=expin)
    expout = '0: hello\n1: hello\n2: hello\n3: hello\n'
    rv = mpdtest.run(cmd="mpiexec%s -l -s all -n 4 cat -" % (PYEXT),
                     expOut=expout,chkOut=1,expIn=expin)
    expout = '0: hello\n2: hello\n3: hello\n'
    rv = mpdtest.run(cmd="mpiexec%s -l -s 0,2-3 -n 4 cat -" % (PYEXT),
                     expOut=expout,chkOut=1,expIn=expin)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1007]:
    print "TEST -1"
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = '%s\n' % (clusterHosts[0])
        rv = mpdtest.run(cmd="mpiexec%s -1 -n 1 /bin/hostname" % (PYEXT), expOut=expout,chkOut=1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[1008]:
    print "TEST -ifhn"
    # not a particularly good test; you can hang/fail with an invalid ifhn
    # ifhn is not very useful for mpiexec since mpd can fill it in as needed
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = 'hello\n'
        rv = mpdtest.run(cmd="mpiexec%s -ifhn 127.0.0.1 -n 1 /bin/echo hello" % (PYEXT),
                         expOut=expout,chkOut=1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[1009]:
    print "TEST -n"
    NMPDS = 1
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
        expout = '0: hello\n1: bye\n'
        mpdtest.run(cmd="mpiexec%s -l -n 1 echo hello : -n 1 echo bye" % (PYEXT),chkOut=1,expOut=expout)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1010]:
    print "TEST -wdir"
    # not a particularly good test; you can hang/fail with an invalid ifhn
    # ifhn is not very useful for mpiexec since mpd can fill it in as needed
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = '/tmp\n/tmp\n'
        rv = mpdtest.run(cmd="mpiexec%s -wdir /tmp -n 2 /bin/pwd" % (PYEXT),
                         expOut=expout,chkOut=1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[1011]:
    print "TEST -path"
    # not a particularly good test; you can hang/fail with an invalid ifhn
    # ifhn is not very useful for mpiexec since mpd can fill it in as needed
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = '/tmp:/bin\n/tmp:/bin\n'
        mpdtest.run(cmd="mpiexec%s -path /tmp:/bin -n 2 /usr/bin/printenv | grep PATH" % (PYEXT),
                    expOut=expout,chkOut=1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[1012]:
    print "TEST -host"
    NMPDS = 5
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = '%s\n' % clusterHosts[3]
        rv = mpdtest.run(cmd="mpiexec%s -n 1 -host %s /bin/hostname" % (PYEXT,clusterHosts[3]),
                         expOut=expout,chkOut=1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[1013]:
    print "TEST -soft"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = 'hello\nhello\nhello\nhello\nhello\n'  # 5 times
    rv = mpdtest.run(cmd="mpiexec%s -n 9 -soft 1:5:2 /bin/echo hello" % (PYEXT),
                     expOut=expout,chkOut=1)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1014]:
    print "TEST -envall (the default)"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = 'BAR\n'
    os.environ['FOO'] = 'BAR'
    rv = mpdtest.run(cmd="mpiexec%s -n 1 -envall sh -c '/bin/echo $FOO'" % (PYEXT),
                     expOut=expout,chkOut=1)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.unsetenv('FOO')   # get rid of this for later tests

if testsToRun[1015]:
    print "TEST -envnone"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = '\n'
    os.environ['FOO'] = ''
    rv = mpdtest.run(cmd="mpiexec%s -n 1 -envnone sh -c '/bin/echo $FOO'" % (PYEXT),
                     expOut=expout,chkOut=1)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.unsetenv('FOO')   # get rid of this for later tests

if testsToRun[1016]:
    print "TEST -env"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = 'BAR\n'
    rv = mpdtest.run(cmd="mpiexec%s -n 1 -env FOO BAR sh -c '/bin/echo $FOO'" % (PYEXT),
                     expOut=expout,chkOut=1)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1017]:
    print "TEST -envlist"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    os.environ['FOO'] = 'BAR'
    os.environ['RMB'] = 'ZZZ'
    expout = 'BAR ZZZ\n'
    rv = mpdtest.run(cmd="mpiexec%s -n 1 -envlist FOO,RMB sh -c '/bin/echo $FOO $RMB'" % (PYEXT),
                     expOut=expout,chkOut=1)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.unsetenv('FOO')   # get rid of this for later tests
    os.unsetenv('RMB')   # get rid of this for later tests

if testsToRun[1018]:
    print "TEST -gn"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = 'hello\nhello\nbye\nbye\n'
    rv = mpdtest.run(cmd="mpiexec%s -gn 2 /bin/echo hello : /bin/echo bye" % (PYEXT),
                     expOut=expout,chkOut=1)
    rv = mpdtest.run(cmd="mpiexec%s -gn 2 : /bin/echo hello : /bin/echo bye" % (PYEXT),
                     expOut=expout,chkOut=1)
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[1019]:
    print "TEST -gexec"
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = '%s\n%s\n' % (socket.gethostname(),clusterHosts[0])
        rv = mpdtest.run(cmd="mpiexec%s -gexec hostname : -n 1 : -n 1" % (PYEXT),
                         expOut=expout,chkOut=1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[1020]:
    print "TEST -genvlist"
    NMPDS = 2
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        os.environ['FOO'] = 'BAR'
        os.environ['RMB'] = 'ZZZ'
        expout = 'BAR ZZZ\n'
        rv = mpdtest.run(cmd="mpiexec%s -genvlist FOO,RMB : sh -c '/bin/echo $FOO $RMB'" % (PYEXT),
                         expOut=expout,chkOut=1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)
        os.unsetenv('FOO')   # get rid of this for later tests
        os.unsetenv('RMB')   # get rid of this for later tests


if 1 in testsToRun[2000:3000]:
    print "MPI tests-------------------------------------------"
    if not MPI_srcdir:
        print "    skipping all MPI tests because no MPI_srcidr is specified"
        for i in range(2000,3000): testsToRun[i] = 0

if testsToRun[2000]:
    print "TEST cpi"
    NMPDS = 1
    if not os.access("%s/examples/cpi" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/examples ; make cpi" % (MPI_srcdir) )
    mpdtest = MPDTest()
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = ['Process 0 of 3','Process 1 of 3','Process 2 of 3']
    rv = mpdtest.run(cmd="mpiexec%s -n 3 %s/examples/cpi" % (PYEXT,MPI_srcdir),
                     grepOut=1, expOut=expout )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[2001]:
    print "TEST spawn1"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/test/mpi/spawn/spawn1" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/test/util ; make" % (MPI_srcdir) )
        os.system("cd %s/test/mpi/spawn ; make spawn1" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = ['No Errors']
    os.system('cp %s/test/mpi/spawn/spawn1 .' % (MPI_srcdir))
    rv = mpdtest.run(cmd="mpiexec%s -n 1 ./spawn1" % (PYEXT),  # -n 1
                     grepOut=1, expOut=expout )
    os.system('rm -f spawn1')
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[2002]:
    print "TEST spawn2"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/test/mpi/spawn/spawn2" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/test/util ; make" % (MPI_srcdir) )
        os.system("cd %s/test/mpi/spawn ; make spawn2" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    os.system('cp %s/test/mpi/spawn/spawn2 .' % (MPI_srcdir))
    expout = ['No Errors']
    rv = mpdtest.run(cmd="mpiexec%s -n 1 ./spawn2" % (PYEXT),  # -n 1
                     grepOut=1, expOut=expout )
    os.system('rm -f spawn2')
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[2003]:
    print "TEST spawnmult2"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/test/mpi/spawn/spawnmult2" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/test/util ; make" % (MPI_srcdir) )
        os.system("cd %s/test/mpi/spawn ; make spawnmult2" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    os.system('cp %s/test/mpi/spawn/spawnmult2 .' % (MPI_srcdir))
    expout = ['No Errors']
    rv = mpdtest.run(cmd="mpiexec%s -n 2 ./spawnmult2" % (PYEXT),  # -n 2
                     grepOut=1, expOut=expout )
    os.system('rm -f spawnmult2')
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    
if testsToRun[2004]:
    print "TEST spawnargv"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/test/mpi/spawn/spawnargv" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/test/util ; make" % (MPI_srcdir) )
        os.system("cd %s/test/mpi/spawn ; make spawnargv" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    os.system('cp %s/test/mpi/spawn/spawnargv .' % (MPI_srcdir))
    expout = ['No Errors']
    rv = mpdtest.run(cmd="mpiexec%s -n 1 ./spawnargv" % (PYEXT),  # -n 2
                     grepOut=1, expOut=expout )
    os.system('rm -f spawnargv')
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[2005]:
    print "TEST spawnintra"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/test/mpi/spawn/spawnintra" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/test/util ; make" % (MPI_srcdir) )
        os.system("cd %s/test/mpi/spawn ; make spawnintra" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    os.system('cp %s/test/mpi/spawn/spawnintra .' % (MPI_srcdir))
    expout = ['No Errors']
    rv = mpdtest.run(cmd="mpiexec%s -n 1 ./spawnintra" % (PYEXT),  # -n 2
                     grepOut=1, expOut=expout )
    os.system('rm -f spawnintra')
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    
if testsToRun[2006]:
    print "TEST namepub"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/test/mpi/spawn/namepub" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/test/util ; make" % (MPI_srcdir) )
        os.system("cd %s/test/mpi/spawn ; make namepub" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = ['No Errors']
    os.system('cp %s/test/mpi/spawn/namepub .' % (MPI_srcdir))
    rv = mpdtest.run(cmd="mpiexec%s -n 1 ./namepub" % (PYEXT),  # -n 2
                     grepOut=1, expOut=expout )
    os.system('rm -f namepub')
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    
if testsToRun[2007]:
    print "TEST concurrent_spawns"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/test/mpi/spawn/concurrent_spawns" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/test/util ; make" % (MPI_srcdir) )
        os.system("cd %s/test/mpi/spawn ; make concurrent_spawns" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = ['No Errors']
    os.system('cp %s/test/mpi/spawn/concurrent_spawns .' % (MPI_srcdir))
    rv = mpdtest.run(cmd="mpiexec%s -n 1 ./concurrent_spawns" % (PYEXT),  # -n 2
                     grepOut=1, expOut=expout )
    os.system('rm -f concurrent_spawns')
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[2008]:
    print "TEST singleton init (cpi)"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/examples/cpi" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/examples ; make cpi" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    expout = ['Process 0 of 1','pi is approximately 3']
    rv = mpdtest.run(cmd="%s/examples/cpi" % (MPI_srcdir), grepOut=1, expOut=expout )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[2009]:
    print "TEST bnr (mpich1-compat using cpi)"
    if not MPI_1_dir:
        print "    skipping; no mpich1 dir specified"
    else:
        NMPDS = 1
        mpdtest = MPDTest()
        if not os.access("%s/examples/cpi" % (MPI_srcdir),os.R_OK):  
            os.system("cd %s/examples ; make cpi" % (MPI_srcdir) )
            os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
            os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
            expout = ['Process 0 on','Process 1 on','Process 2 on','pi is approximately 3']
            rv = mpdtest.run(cmd="mpiexec%s -bnr -n 3 %s/examples/cpi" % (PYEXT,MPI_1_dir),
                             grepOut=1, expOut=expout )
            os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )


if 1 in testsToRun[3000:4000]:
    print "console pgm tests---------------------------------------------------"

if testsToRun[3000]:
    print "TEST mpdtrace"
    NMPDS = 3
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        expout = ['%s' % (socket.gethostname()),
                  '%s' % (clusterHosts[0]),
                  '%s' % (clusterHosts[1])]
        rv = mpdtest.run(cmd="mpdtrace%s -l" % (PYEXT), grepOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[3001]:
    print "TEST mpdringtest"
    NMPDS = 3
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        nLoops = 100
        expout = ['time for %d loops' % (nLoops) ]
        rv = mpdtest.run(cmd="mpdringtest%s %d" % (PYEXT,nLoops), grepOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[3002]:
    print "TEST mpdexit"
    NMPDS = 3
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        temph = open(HFILE,'w')
        for host in clusterHosts: print >>temph, host
        temph.close()
        os.system("mpdboot%s -f %s -n %d" % (PYEXT,HFILE,NMPDS) )
        linesAsStr = commands.getoutput("mpdtrace%s -l" % (PYEXT) )
        for line in linesAsStr.split('\n'):
            if line.find(clusterHosts[0]) >= 0:
                mpdid = line.split(' ')[0]  # strip off the (ifhn) stuff
                break
        else:
            mpdid = ''
        if mpdid:
            rv = mpdtest.run(cmd="mpdexit%s %s" % (PYEXT,mpdid), chkOut=1 )
        else:
            print 'failed to find %s in mpdtrace for mpdexit' % (clusterHosts[0])
            sys.exit(-1)
        expout = ['%s' % (socket.gethostname()), '%s' % (clusterHosts[1])]
        rv = mpdtest.run(cmd="mpdtrace%s -l" % (PYEXT),
                         grepOut=1, expOut=expout )
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        os.unlink(HFILE)

if testsToRun[3003]:
    print "TEST mpdlistjobs and mpdkilljob"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/src/pm/mpd/infloop" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/src/pm/mpd ; cc -o infloop infloop.c" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    if not os.access("./infloop",os.R_OK):  
       os.system('cp %s/src/pm/mpd/infloop .' % (MPI_srcdir))
    os.system("mpiexec%s -s all -a ralph -n 2 ./infloop -p &" % (PYEXT))  # -p => don't print
    import time    ## give the mpiexec
    time.sleep(2)  ##     time to get going
    expout = ['ralph']
    rv = mpdtest.run(cmd="mpdlistjobs%s" % (PYEXT), grepOut=1, expOut=expout )
    rv = mpdtest.run(cmd="mpdkilljob%s -a ralph" % (PYEXT), chkOut=0 )
    expout = ''
    rv = mpdtest.run(cmd="mpdlistjobs%s" % (PYEXT), chkOut=1, expOut=expout )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("killall -q infloop")  ## just to be safe
    os.system('rm -f infloop')
    
if testsToRun[3004]:
    print "TEST mpdsigjob"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/src/pm/mpd/infloop" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/src/pm/mpd ; cc -o infloop infloop.c" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    if not os.access("./infloop",os.R_OK):  
        os.system('cp %s/src/pm/mpd/infloop .' % (MPI_srcdir))
    os.system("mpiexec%s -s all -a ralph -n 2 ./infloop -p &" % (PYEXT))  # -p => don't print
    import time    ## give the mpiexec
    time.sleep(2)  ##     time to get going
    expout = ['ralph']
    rv = mpdtest.run(cmd="mpdlistjobs%s" % (PYEXT), grepOut=1, expOut=expout )
    rv = mpdtest.run(cmd="mpdsigjob%s INT -a ralph -g" % (PYEXT), chkOut=0 )
    expout = ''
    rv = mpdtest.run(cmd="mpdlistjobs%s" % (PYEXT), chkOut=1, expOut=expout )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system('rm -f infloop')
    

if 1 in testsToRun[8000:9000]:
    print "misc tests---------------------------------------------------"

if testsToRun[8000]:
    print "TEST ^C to mpiexec"
    NMPDS = 1
    mpdtest = MPDTest()
    if not os.access("%s/src/pm/mpd/infloop" % (MPI_srcdir),os.R_OK):  
        os.system("cd %s/src/pm/mpd ; cc -o infloop infloop.c" % (MPI_srcdir) )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("mpdboot%s -n %d" % (PYEXT,NMPDS) )
    if not os.access("./infloop",os.R_OK):  
        os.system('cp %s/src/pm/mpd/infloop .' % (MPI_srcdir))
    import popen2
    runner = popen2.Popen4("mpiexec%s -n 2 ./infloop -p" % (PYEXT))  # -p => don't print
    import time    ## give the mpiexec
    time.sleep(2)  ##     time to get going
    os.system("kill -INT %d" % (runner.pid) )  # simulate user ^C
    expout = ''
    rv = mpdtest.run(cmd="mpdlistjobs%s" % (PYEXT), chkOut=1, expOut=expout )
    os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
    os.system("killall -q infloop")  ## just to be safe
    os.system('rm -f infloop')

if testsToRun[8001]:
    print "TEST re-knit a ring"
    NMPDS = 3
    if NMPDS > len(clusterHosts)+1:
        print "    skipping; too few hosts"
    else:
        mpdtest = MPDTest()
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )
        import popen2, time
        mpd1 = popen2.Popen4("mpd%s -l 12345" % (PYEXT))
        time.sleep(2)
        mpd2 = popen2.Popen4("mpd%s -n -h %s -p 12345" % (PYEXT,socket.gethostname()) )
        mpd3 = popen2.Popen4("mpd%s -n -h %s -p 12345" % (PYEXT,socket.gethostname()) )
        time.sleep(2)
        rv = mpdtest.run(cmd="mpdtrace%s" % (PYEXT), chkOut=0 )
        if len(rv['OUT']) != NMPDS:
            print "a: unexpected number of lines of output from mpdtrace", rv['OUT']
            sys.exit(-1)
        hostname = socket.gethostname()
        for line in rv['OUT']:
            if line.find(hostname) < 0:
                print "a: bad lines in output of mpdtrace", rv['OUT']
                sys.exit(-1)
        os.system("kill -9 %d" % (mpd3.pid) )
        time.sleep(1)
        rv = mpdtest.run(cmd="mpdtrace%s" % (PYEXT), chkOut=0 )
        if len(rv['OUT']) != NMPDS-1:
            print "b: unexpected number of lines of output from mpdtrace", rv['OUT']
            sys.exit(-1)
        hostname = socket.gethostname()
        for line in rv['OUT']:
            if line.find(hostname) < 0:
                print "b: bad lines in output of mpdtrace", rv['OUT']
                sys.exit(-1)
        os.system("mpdallexit%s 1> /dev/null 2> /dev/null" % (PYEXT) )

if testsToRun[8002]:
    print "TEST installing mpd as user"
    if not os.access("%s/src/pm/mpd/mpdroot" % (MPI_srcdir),os.X_OK):
        os.system("cd %s/src/pm/mpd ; ./configure ; make")
    os.system("cd %s/src/pm/mpd ; make prefix=%s install" % (MPI_srcdir,USER_INSTALL_DIR) )
    print "TEST the mpd installed as user ; run mpdtrace as user"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("%s/bin/mpdallexit%s 1> /dev/null 2> /dev/null" % (USER_INSTALL_DIR,PYEXT) )
    os.system("%s/bin/mpdboot%s -n %d" % (USER_INSTALL_DIR,PYEXT,NMPDS) )
    expout = ['%s' % (socket.gethostname())]
    rv = mpdtest.run(cmd="%s/bin/mpdtrace%s -l" % (USER_INSTALL_DIR,PYEXT), grepOut=1, expOut=expout )
    os.system("%s/bin/mpdallexit%s 1> /dev/null 2> /dev/null" % (USER_INSTALL_DIR,PYEXT) )


if 1 in testsToRun[9000:10000]:
    print "root tests---------------------------------------------------"
    import popen2
    runner = popen2.Popen4("sudo echo hello")
    sout_serr = runner.fromchild
    line = sout_serr.readline()
    if line.find('hello') < 0:
        print "    ** you do not have sudo capability; terminating\n"
        sys.exit(-1)

if testsToRun[9000]:
    os.environ['MPD_CON_EXT'] = ""    # get rid of this for root testing
    print "TEST installing mpd as root"
    if not os.access("%s/src/pm/mpd/mpdroot" % (MPI_srcdir),os.X_OK):
        os.system("cd %s/src/pm/mpd ; ./configure ; make")
    os.system("cd %s/src/pm/mpd ; sudo make prefix=%s install" % (MPI_srcdir,ROOT_INSTALL_DIR) )
    print "TEST mpd as root ; mpdtrace as user"
    NMPDS = 1
    mpdtest = MPDTest()
    os.system("sudo %s/bin/mpdallexit%s 1> /dev/null 2> /dev/null" % (ROOT_INSTALL_DIR,PYEXT) )
    os.system("sudo %s/bin/mpd%s -d" % (ROOT_INSTALL_DIR,PYEXT) )  # not using boot here
    import time
    time.sleep(2)
    os.environ['MPD_USE_ROOT_MPD'] = '1'
    # os.system("%s/bin/mpdtrace%s -l" % (ROOT_INSTALL_DIR,PYEXT))
    expout = ['%s' % (socket.gethostname())]
    rv = mpdtest.run(cmd="%s/bin/mpdtrace%s -l" % (ROOT_INSTALL_DIR,PYEXT), grepOut=1, expOut=expout )

    print "TEST that user cannot remove files owned by root"
    os.system("sudo touch /tmp/testroot")
    expout = ['cannot remove']
    rv = mpdtest.run(cmd="%s/bin/mpiexec%s -n 1 /bin/rm -f /tmp/testroot" % (ROOT_INSTALL_DIR,PYEXT),
                     grepOut=1, expOut=expout )
    os.system("sudo rm -f /tmp/testroot")
    os.system("sudo %s/bin/mpdallexit%s 1> /dev/null 2> /dev/null" % (ROOT_INSTALL_DIR,PYEXT) )



#!/usr/bin/env python
#
#   (C) 2001 by Argonne National Laboratory.
#       See COPYRIGHT in top-level directory.
#

"""
The following mpd commands are available.  For usage of any specific one,
invoke it with the single argument --help .

mpd           start an mpd daemon
mpdtrace      show all mpd's in ring
mpdboot       start a ring of daemons all at once
mpdringtest   test how long it takes for a message to circle the ring 
mpdexit       remove one mpd from the ring
mpdallexit    take down all daemons in ring
mpdcleanup    repair local Unix socket if ring crashed badly
mpdlistjobs   list processes of jobs (-a or --all: all jobs for all users)
mpdkilljob    kill all processes of a single job
mpdsigjob     deliver a specific signal to the application processes of a job
mpiexec       start a parallel job

Each command can be invoked with the --help argument, which prints usage
information for the command without running it.
"""
from time import ctime
__author__ = "Ralph Butler and Rusty Lusk"
__date__ = ctime()
__version__ = "$Revision: 1.8 $"
__credits__ = ""


if __name__ == '__main__':
    print __doc__



#!/usr/bin/env python
#
#   (C) 2001 by Argonne National Laboratory.
#       See COPYRIGHT in top-level directory.
#

"""
This program is not to be executed from the command line.  It is 
exec'd by mpdman to support mpigdb.
"""

# workaround to suppress deprecated module warnings in python2.6
# see https://trac.mpich.org/projects/mpich/ticket/362 for tracking
import warnings
warnings.filterwarnings('ignore', '.*the popen2 module is deprecated.*', DeprecationWarning)

from time import ctime
__author__ = "Ralph Butler and Rusty Lusk"
__date__ = ctime()
__version__ = "$Revision: 1.17 $"
__credits__ = ""


from sys    import argv, exit, stdin, stdout, stderr
from os     import kill, getpid, write, strerror
from popen2 import Popen4
from signal import signal, SIGUSR1, SIGINT, SIGKILL
from errno  import EINTR
from select import select, error
from re     import findall, sub
from mpdlib import mpd_set_my_id, mpd_print

global appPid, gdbPID


def sig_handler(signum,frame):
    global appPid, gdbPID
    if signum == SIGINT:
        try:
            kill(appPid,SIGINT)
        except:
            pass
    elif signum == SIGUSR1:
        try:
            kill(gdbPid,SIGKILL)
        except:
            pass
        try:
            kill(appPid,SIGKILL)
        except:
            pass


if __name__ == '__main__':    # so I can be imported by pydoc
    signal(SIGINT,sig_handler)
    signal(SIGUSR1,sig_handler)
    mpd_set_my_id('mpdgdbdrv')
    
    ## mpd_print(1,"RMB:GDBDRV: ARGS=%s" % argv)
    if argv[1] == '-attach':
        gdb_args = '%s %s' % (argv[2],argv[3])  # userpgm and userpid
    else:
        if len(argv) > 2:
            mpd_print(1, "when using gdb, pass cmd-line args to user pgms via the 'run' cmd")
            exit(-1)
        gdb_args = argv[1]
    gdb_info = Popen4('gdb -q %s' % (gdb_args), 0 )
    gdbPid = gdb_info.pid
    # print "PID=%d GDBPID=%d" % (getpid(),gdbPid) ; stdout.flush()
    gdb_sin = gdb_info.tochild
    gdb_sin_fileno = gdb_sin.fileno()
    gdb_sout_serr = gdb_info.fromchild
    gdb_sout_serr_fileno = gdb_sout_serr.fileno()
    write(gdb_sin_fileno,'set prompt (gdb)\\n\n')
    gdb_line = gdb_sout_serr.readline() 
    # check if gdb reports any errors
    if findall(r'.*: No such file or directory.',gdb_line) != []:
        print gdb_line, ; stdout.flush()
        exit(-1)
    mpd_print(0000, "LINE1=|%s|" % (gdb_line.rstrip()))
    write(gdb_sin_fileno,'set confirm off\n')
    gdb_line = gdb_sout_serr.readline() 
    mpd_print(0000, "LINE2=|%s|" % (gdb_line.rstrip()))
    write(gdb_sin_fileno,'handle SIGUSR1 nostop noprint\n')
    gdb_line = gdb_sout_serr.readline() 
    mpd_print(0000, "LINE3=|%s|" % (gdb_line.rstrip()))
    write(gdb_sin_fileno,'handle SIGPIPE nostop noprint\n')
    gdb_line = gdb_sout_serr.readline() 
    mpd_print(0000, "LINE4=|%s|" % (gdb_line.rstrip()))
    write(gdb_sin_fileno,'set confirm on\n')
    gdb_line = gdb_sout_serr.readline() 
    mpd_print(0000, "LINE5=|%s|" % (gdb_line.rstrip()))
    write(gdb_sin_fileno,'echo hi1\n')
    gdb_line = gdb_sout_serr.readline() 
    # mpd_print(0000, "LINE6=|%s|" % (gdb_line.rstrip()))
    # gdb_line = ''
    while not gdb_line.startswith('hi1'):
        gdb_line = gdb_sout_serr.readline() 
        mpd_print(0000, "LINEx=|%s|" % (gdb_line.rstrip()))

    if argv[1] != '-attach':
        write(gdb_sin_fileno,'b main\n')
        gdb_line = ''
        while not gdb_line.startswith('Breakpoint'):
            try:
                (readyFDs,unused1,unused2) = select([gdb_sout_serr_fileno],[],[],10)
            except error, data:
                if data[0] == EINTR:    # interrupted by timeout for example
                    continue
                else:
                    print 'mpdgdb_drv: main loop: select error: %s' % strerror(data[0])
            if not readyFDs:
                mpd_print(1, 'timed out waiting for initial Breakpoint response')
                exit(-1)
            gdb_line = gdb_sout_serr.readline()  # drain breakpoint response
            gdb_line = gdb_line.strip()
            mpd_print(0000, "gdb_line=|%s|" % (gdb_line.rstrip()))
        if not gdb_line.startswith('Breakpoint'):
            mpd_print(1, 'expecting "Breakpoint", got :%s:' % (gdb_line) )
            exit(-1)
        gdb_line = gdb_sout_serr.readline()  # drain prompt
        mpd_print(0000, "gdb_line=|%s|" % (gdb_line.rstrip()))
        if not gdb_line.startswith('(gdb)'):
            mpd_print(1, 'expecting "(gdb)", got :%s:' % (gdb_line) )
            exit(-1)
    
    print '(gdb)\n', ; stdout.flush()    # initial prompt to user
    
    user_fileno = stdin.fileno()
    while 1:
        try:
            (readyFDs,unused1,unused2) = select([user_fileno,gdb_sout_serr_fileno],[],[],1)
        except error, data:
            if data[0] == EINTR:    # interrupted by timeout for example
                continue
            else:
                mpd_print(1, 'mpdgdbdrv: main loop: select error: %s' % strerror(data[0]))
        # print "READY=", readyFDs ; stdout.flush()
        for readyFD in readyFDs:
            if readyFD == gdb_sout_serr_fileno:
                gdb_line = gdb_sout_serr.readline()
                if not gdb_line:
                    print "MPIGDB ENDING" ; stdout.flush()
                    exit(0)
                # print "LINE |%s|" % (gdb_line.rstrip()) ; stdout.flush()
                print gdb_line, ; stdout.flush()
            elif readyFD == user_fileno:
                user_line = stdin.readline()
                # print "USERLINE=", user_line, ; stdout.flush()
                if not user_line:
                    mpd_print(1, 'mpdgdbdrv: problem: expected user input but got none')
                    exit(-1)
                if user_line.startswith('r'):
                    write(gdb_sin_fileno,'show prompt\n')
                    gdb_line = gdb_sout_serr.readline()
                    gdb_prompt = findall(r'Gdb\'s prompt is "(.+)"\.',gdb_line)
                    if gdb_prompt == []:
                        mpd_print(1, 'expecting gdb\'s prompt, got :%s:' % (gdb_line))
                        exit(-1)
                    gdb_prompt = gdb_prompt[0]
                    # cut everything after first escape character (including it)
                    p = gdb_prompt.find("\\")
                    if p > 0:
                        gdb_prompt = gdb_prompt[0:p]
                    gdb_line = gdb_sout_serr.readline() # drain one line

                    write(gdb_sin_fileno,'show confirm\n')
                    gdb_line = gdb_sout_serr.readline()
                    gdb_confirm = findall(r'Whether to confirm potentially dangerous operations is (on|off)\.',gdb_line)
                    if gdb_confirm == []:
                        mpd_print(1, 'expecting gdb\'s confirm state, got :%s:' % (gdb_line))
                        exit(-1)
                    gdb_confirm = gdb_confirm[0]
                    gdb_line = gdb_sout_serr.readline() # drain one line

                    # set confirm to 'on' to get 'Starting program' message
                    write(gdb_sin_fileno,'set confirm on\n')
                    gdb_line = gdb_sout_serr.readline()

                    # we have already set breakpoint 1 in main
                    write(gdb_sin_fileno,user_line)
                    # ignore any warnings befor starting msg
                    while 1:
                        gdb_line = gdb_sout_serr.readline()  # drain one line
                        if not gdb_line.startswith('warning:'):
                            break
                        else:
                            print gdb_line, ; stdout.flush()
                    # drain starting msg
                    if not gdb_line.startswith('Starting program'):
                        mpd_print(1, 'expecting "Starting program", got :%s:' % \
                                  (gdb_line))
                        exit(-1)
                    while 1:    # drain to a prompt
                        gdb_line = gdb_sout_serr.readline()  # drain one line
                        if gdb_line.startswith(gdb_prompt):
                            break
                    # try to get the pid
                    write(gdb_sin_fileno,'info pid\n')  # macosx
                    gdb_line = gdb_sout_serr.readline().lstrip()
                    if gdb_line.find('process ID') >= 0:  # macosx
                        appPid = findall(r'.* has process ID (\d+)',gdb_line)
                        appPid = int(appPid[0])
                    else:
                        while 1:    # drain to a prompt
                            gdb_line = gdb_sout_serr.readline()  # drain one line
                            if gdb_line.startswith(gdb_prompt):
                                break
                        write(gdb_sin_fileno,'info program\n')
                        gdb_line = gdb_sout_serr.readline().lstrip()
                        if gdb_line.startswith('Using'):
                            if gdb_line.find('process') >= 0:
                                appPid = findall(r'Using .* image of child process (\d+)',gdb_line)
                            elif gdb_line.find('Thread') >= 0:  # solaris
                                appPid = findall(r'Using .* image of child .* \(LWP (\d+)\).',gdb_line)
                            else:
                                mpd_print(1, 'expecting process or thread line, got :%s:' % \
                                          (gdb_line))
                                exit(-1)
                            appPid = int(appPid[0])
                        else:
                            mpd_print(1, 'expecting line with "Using"; got :%s:' % (gdb_line))
                            exit(-1)
                    while 1:    # drain to a prompt
                        gdb_line = gdb_sout_serr.readline()  # drain one line
                        if gdb_line.startswith(gdb_prompt):
                            break
                    write(gdb_sin_fileno,'c\n')
                    # set confirm back to original state
                    write(gdb_sin_fileno,'set confirm %s\n' % (gdb_confirm))
                else:
                    write(gdb_sin_fileno,user_line)



#!/usr/bin/env python
#
#   (C) 2001 by Argonne National Laboratory.
#       See COPYRIGHT in top-level directory.
#

import sys, os, socket

manPort = int(sys.argv[1])
manSock = socket.socket()
manSock.connect((socket.gethostname(),manPort))
msg = manSock.recv(3)    # 'go\n'
if not msg.startswith('go'):
    print 'mpscliwrap: invalid go msg from man'
    sys.exit(-1)

try:     max_fds = os.sysconf('SC_OPEN_MAX')
except:  max_fds = 1024
for fd in range(3,max_fds):
    try:     os.close(fd)
    except:  pass
# can NOT use syslog (including inside an mpd_print) below here without
# doing a new openlog because any syslog fd would have been closed

os.execvpe(sys.argv[2],sys.argv[3:],os.environ)    # client



#!/usr/bin/env python
#
#   (C) 2001 by Argonne National Laboratory.
#       See COPYRIGHT in top-level directory.
#

"""
mpdcheck

This script is a work in progress and may change frequently as we work
with users and gain additional insights into how to improve it.

This script prints useful information about the host on which it runs.
It is here to help us help users detect problems with configurations of
their computers.  For example, some computers are configured to think
of themselves simply as 'localhost' with 127.0.0.1 as the IP address.
This might present problems if a process on that computer wishes to
identify itself by host and port to a process on another computer.
The process on the other computer would try to contact 'localhost'.

If you are having problems running parallel jobs via mpd on one or more
hosts, you might try running this script once on each of those hosts.

Any output with *** at the beginning indicates a potential problem
that you may have to resolve before being able to run parallel jobs
via mpd.

For help:
    mpdcheck -h (or --help)
        prints this message

In the following modes, the -v (verbose) option provides info about what
mpdcheck is doing; the -l (long messages) option causes long informational
messages to print in situations where problems are spotted.

The three major modes of operation for this program are:

    mpdcheck
        looks for config problems on 'this' host; prints as nec

    mpdcheck -pc
        print config info about 'this' host, e.g. contents of /etc/hosts, etc.

    mpdcheck -f some_file [-ssh]
        prints info about 'this' host and locatability info about the ones
        listed in some_file as well (note the file might be mpd.hosts);
        the -ssh option can be used in conjunction with the -f option to
        cause ssh tests to be run to each remote host

    mpdcheck -s
        runs this program as a server on one host
    mpdcheck -c server_host server_port
        runs a client on another (or same) host; connects to the specifed
        host/port where you previously started the server
"""

# workaround to suppress deprecated module warnings in python2.6
# see https://trac.mpich.org/projects/mpich/ticket/362 for tracking
import warnings
warnings.filterwarnings('ignore', '.*the popen2 module is deprecated.*', DeprecationWarning)

from time import ctime
__author__ = "Ralph Butler and Rusty Lusk"
__date__ = ctime()
__version__ = "$Revision: 1.19 $"
__credits__ = ""

import re

from  sys      import argv, exit, stdout
from  os       import path, kill, system
from  signal   import SIGKILL
from  socket   import gethostname, getfqdn, gethostbyname_ex, gethostbyaddr, socket
from  popen2   import Popen3
from  select   import select, error
from  commands import getoutput


if __name__ == '__main__':    # so I can be imported by pydoc
    do_ssh = 0
    fullDirName = path.abspath(path.split(argv[0])[0])  # normalize
    hostsFromFile = []
    verbose = 0
    long_messages = 0
    argidx = 1
    while argidx < len(argv):
        if argv[argidx] == '-h'  or argv[argidx] == '--help':
            print __doc__
            exit(0)
        elif argv[argidx] == '-s':
            lsock = socket()
            lsock.bind(('',0)) # anonymous port
            lsock.listen(5)
            print "server listening at INADDR_ANY on: %s %s" % (gethostname(),lsock.getsockname()[1])
            stdout.flush()
            (tsock,taddr) = lsock.accept()
            print "server has conn on %s from %s" % (tsock,taddr)
            msg = tsock.recv(64)
            if not msg:
                print "*** server failed to recv msg from client"
            else:
                print "server successfully recvd msg from client: %s" % (msg)
            tsock.sendall('ack_from_server_to_client')
            tsock.close()
            lsock.close()
            exit(0)
        elif argv[argidx] == '-c':
            sock = socket()
            sock.connect((argv[argidx+1],int(argv[argidx+2])))  # note double parens
            sock.sendall('hello_from_client_to_server')
            msg = sock.recv(64)
            if not msg:
                print "*** client failed to recv ack from server"
            else:
                print "client successfully recvd ack from server: %s" % (msg)
                stdout.flush()
            sock.close()
            exit(0)
        elif argv[argidx] == '-pc':
            print "--- print results of: gethostbyname_ex(gethostname())"
            print gethostbyname_ex(gethostname())
            print "--- try to run /bin/hostname"
            linesAsStr = getoutput("/bin/hostname")
            print linesAsStr
            print "--- try to run uname -a"
            linesAsStr = getoutput("/bin/uname -a")
            print linesAsStr
            print "--- try to print /etc/hosts"
            linesAsStr = getoutput("/bin/cat /etc/hosts")
            print linesAsStr
            print "--- try to print /etc/resolv.conf"
            linesAsStr = getoutput("/bin/cat /etc/resolv.conf")
            print linesAsStr
            print "--- try to run /sbin/ifconfig -a"
            linesAsStr = getoutput("/sbin/ifconfig -a")
            print linesAsStr
            print "--- try to print /etc/nsswitch.conf"
            linesAsStr = getoutput("/bin/cat /etc/nsswitch.conf")
            print linesAsStr
            exit(0)
        elif argv[argidx] == '-v':
            verbose = 1
            argidx += 1
        elif argv[argidx] == '-l':
            long_messages = 1
            argidx += 1
        elif argv[argidx] == '-f':
            try:
                hostsFile = open(argv[argidx+1])
            except:
                print 'unable to open file ', argv[argidx+1]
                exit(-1)
            for line in hostsFile:
                line = line.rstrip()
                if not line  or  line[0] == '#':
                    continue
                splitLine = re.split(r'\s+',line)
                host = splitLine[0]
                if ':' in host:
                    (host,ncpus) = host.split(':')
                hostsFromFile.append(host)
            argidx += 2
        elif argv[argidx] == '-ssh':
            do_ssh = 1
            argidx += 1
        else:
            print 'unrecognized arg:', argv[argidx]
            exit(0)
    
    
    # See if we can do gethostXXX, etc. for this host
    if verbose:
        print 'obtaining hostname via gethostname and getfqdn'
    uqhn1 = gethostname()
    fqhn1 = getfqdn()
    if verbose:
        print "gethostname gives ", uqhn1
        print "getfqdn gives ", fqhn1
    if verbose:
        print 'checking out unqualified hostname; make sure is not "localhost", etc.'
    if uqhn1.startswith('localhost'):
        if long_messages:
            msg = """
            **********
            The unqualified hostname seems to be localhost. This generally
            means that the machine's hostname is not set. You may change
            it by using the 'hostname' command, e.g.:
                hostname mybox1
            However, this will not remain after a reboot. To do this, you
            will need to consult the operating system's documentation. On
            Debian Linux systems, this can be done by:
                echo "mybox1" > /etc/hostname
            **********
            """
        else:
            msg = "*** the uq hostname seems to be localhost"
        print msg.strip().replace('        ','')
    elif uqhn1 == '':
        if long_messages:
            msg = """
            **********
            The unqualified hostname seems to be blank. This generally
            means that the machine's hostname is not set. You may change
            it by using the 'hostname' command, e.g.:
                hostname mybox1
            However, this will not remain after a reboot. To do this, you
            will need to consult the operating system's documentation. On
            Debian Linux systems, this can be done by:
                echo "mybox1" > /etc/hostname
            **********
            """
        else:
            msg = "*** the uq hostname seems to be localhost"
        print msg.replace('        ','')
    if verbose:
        print 'checking out qualified hostname; make sure is not "localhost", etc.'
    if fqhn1.startswith('localhost'):
        if long_messages:
            msg = """
            **********
            Your fully qualified hostname seems to be set to 'localhost'.
            This generally means that your machine's /etc/hosts file contains a line
            similar to this:
                127.0.0.1 mybox1 localhost.localdomain localhost
            You probably want to remove your hostname from this line and place it on
            a line by itself with your ipaddress, like this:
                $ipaddr mybox1
            **********
            """
        else:
            msg =  "*** the fq hostname seems to be localhost"
        print msg.rstrip().replace('        ','')
    elif fqhn1 == '':
        if long_messages:
            msg = """
            **********
            Your fully qualified hostname seems to be blank.
            **********
            """
        else:
            msg = "*** the fq hostname is blank"
        print msg.replace('        ','')
    
    if verbose:
        print 'obtain IP addrs via qualified and unqualified hostnames;',
        print ' make sure other than 127.0.0.1'
    uipaddr1 = 0
    try:
        ghbnu = gethostbyname_ex(uqhn1)
        if verbose:
            print "gethostbyname_ex: ", ghbnu
        uipaddr1 = ghbnu[2][0]
        if uipaddr1.startswith('127'):
            if long_messages:
                msg = """
                **********
                Your unqualified hostname resolves to 127.0.0.1, which is
                the IP address reserved for localhost. This likely means that
                you have a line similar to this one in your /etc/hosts file:
                127.0.0.1   $uqhn
                This should perhaps be changed to the following:
                127.0.0.1   localhost.localdomain localhost
                **********
                """
            else:
                msg = "*** first ipaddr for this host (via %s) is: %s" % (uqhn1,uipaddr1)
            print msg.replace('            ','')
        try:
            ghbau = gethostbyaddr(uipaddr1)
        except:
            print "*** gethostbyaddr failed for this hosts's IP %s" % (uipaddr1)
    except:
        if long_messages:
            msg = """
            **********
            The system call gethostbyname(3) failed to resolve your
            unqualified hostname, or $uqhn. This can be caused by
            missing info from your /etc/hosts file or your system not
            having correctly configured name resolvers, or by your IP 
            address not existing in resolution services.
            If you run DNS, you may wish to make sure that your
            DNS server has the correct forward A set up for your machine's
            hostname. If you are not using DNS and are only using hosts
            files, please check that a line similar to the one below exists
            in your /etc/hosts file:
                $ipaddr $uqdn
            If you plan to use DNS but you are not sure that it is
            correctly configured, please check that the file /etc/resolv.conf
            contains entries similar to the following:
                nameserver 1.2.3.4
            where 1.2.3.4 is an actual IP of one of your nameservers.
            **********
            """
        else:
            msg = "*** gethostbyname_ex failed for this host %s" % (uqhn1)
        print msg.replace('        ','')
    
    fipaddr1 = 0
    try:
        ghbnf = gethostbyname_ex(fqhn1)
        if verbose:
            print "gethostbyname_ex: ", ghbnf
        fipaddr1 = ghbnf[2][0]
        if fipaddr1.startswith('127'):
            msg = """
            **********
            Your fully qualified hostname resolves to 127.0.0.1, which
            is the IP address reserved for localhost. This likely means
            that you have a line similar to this one in your /etc/hosts file:
                 127.0.0.1   $fqhn
            This should be perhaps changed to the following:
                 127.0.0.1   localhost.localdomain localhost
            **********
            """
        try:
            ghbaf = gethostbyaddr(fipaddr1)
        except:
            print "*** gethostbyaddr failed for this hosts's IP %s" % (uipaddr1)
    except:
        if long_messages:
            msg = """
            **********
            The system call gethostbyname(3) failed to resolve your
            fully qualified hostname, or $fqhn. This can be caused by
            missing info from your /etc/hosts file or your system not
            having correctly configured name resolvers, or by your IP 
            address not existing in resolution services.
            If you run DNS, please check and make sure that your
            DNS server has the correct forward A record set up for your
            machine's hostname. If you are not using DNS and are only using
            hosts files, please check that a line similar to the one below
            exists in your /etc/hosts file:
                $ipaddr $fqhn
            If you intend to use DNS but you are not sure that it is
            correctly configured, please check that the file /etc/resolv.conf
            contains entries similar to the following:
                nameserver 1.2.3.4
            where 1.2.3.4 is an actual IP of one of your nameservers.
            **********
            """
        else:
            msg = "*** gethostbyname_ex failed for host %s" % (fqhn1)
        print msg.replace('        ','')
    
    if verbose:
        print 'checking that IP addrs resolve to same host'
    if uipaddr1 and fipaddr1 and uipaddr1 != fipaddr1:
        msg = """
            **********
            Your fully qualified and unqualified names do not resolve to
            the same IP. This likely means that your DNS domain name is not
            set correctly.  This might be fixed by adding a line similar
            to the following to your /etc/hosts:
                 $ipaddr             $fqhn   $uqdn
            **********
            """
        print msg.replace('        ','')
    
    
    if verbose:
        print 'now do some gethostbyaddr and gethostbyname_ex for machines in hosts file'
    # See if we can do gethostXXX, etc. for hosts in hostsFromFile
    for host in hostsFromFile:
        uqhn2 = host
        fqhn2 = getfqdn(uqhn2)
        uipaddr2 = 0
        if verbose:
            print 'checking gethostbyXXX for unqualified %s' % (uqhn2)
        try:
            ghbnu = gethostbyname_ex(uqhn2)
            if verbose:
                print "gethostbyname_ex: ", ghbnu
            uipaddr2 = ghbnu[2][0]
            try:
                ghbau = gethostbyaddr(uipaddr2)
            except:
                print "*** gethostbyaddr failed for remote hosts's IP %s" % (fipaddr2)
        except:
            print "*** gethostbyname_ex failed for host %s" % (fqhn2)
        if verbose:
            print 'checking gethostbyXXX for qualified %s' % (uqhn2)
        try:
            ghbnf = gethostbyname_ex(fqhn2)
            if verbose:
                print "gethostbyname_ex: ", ghbnf
            fipaddr2 = ghbnf[2][0]
            if uipaddr2  and  fipaddr2 != uipaddr2:
                print "*** ipaddr via uqn (%s) does not match via fqn (%s)" % (uipaddr2,fipaddr2)
            try:
                ghbaf = gethostbyaddr(fipaddr2)
            except:
                print "*** gethostbyaddr failed for remote hosts's IP %s" % (fipaddr2)
        except:
            print "*** gethostbyname_ex failed for host %s" % (fqhn2)
    
    
    # see if we can run /bin/date on remote hosts
    if not do_ssh:
        exit(0)
    
    for host in hostsFromFile:
        cmd = "ssh %s -x -n /bin/echo hello" % (host)
        if verbose:
            print 'trying: %s' % (cmd)
        runner = Popen3(cmd,1,0)
        runout = runner.fromchild
        runerr = runner.childerr
        runin  = runner.tochild
        runpid = runner.pid
        try:
            (readyFDs,unused1,unused2) = select([runout],[],[],9)
        except Exception, data:
            print 'select 1 error: %s ; %s' % ( data.__class__, data)
            exit(-1)
        if len(readyFDs) == 0:
            print '** ssh timed out to %s' % (host)
        line = ''
        failed = 0
        if runout in readyFDs:
            line = runout.readline()
            if not line.startswith('hello'):
                failed = 1
        else:
            failed = 1
        if failed:
            print '** ssh failed to %s' % (host)
            print '** here is the output:'
            if line:
                print line,
            done = 0
            fds = [runout,runerr]
            while not done:
                try:
                    (readyFDs,unused1,unused2) = select(fds,[],[],1)
                except Exception, data:
                    print 'select 2 error: %s ; %s' % ( data.__class__, data)
                    exit(-1)
                if runout in readyFDs:
                    line = runout.readline()
                    if line:
                        print line,
                    else:
                        fds.remove(runout)
                elif runerr in readyFDs:
                    line = runerr.readline()
                    if line:
                        print line,
                    else:
                        fds.remove(runerr)
                else:
                    done = 1
        try:
            kill(runpid,SIGKILL)
            runout.close()
            runerr.close()
            runin.close()
        except:
            pass
        if failed:
            exit(-1)
    
    # see if we can run mpdcheck on remote hosts
    for host in hostsFromFile:
        cmd1 = path.join(fullDirName,'mpdcheck.py') + ' -s'
        if verbose:
            print 'starting server: %s' % (cmd1)
        runner1 = Popen3(cmd1,1,0)
        runout1 = runner1.fromchild
        runerr1 = runner1.childerr
        runin1  = runner1.tochild
        runpid1 = runner1.pid
        try:
            (readyFDs,unused1,unused2) = select([runout1],[],[],9)
        except Exception, data:
            print 'select 3 error: %s ; %s' % ( data.__class__, data)
            exit(-1)
        if len(readyFDs) == 0:
            print '** timed out waiting for local server to produce output'
        line = ''
        failed = 0
        port = 0
        if runout1 in readyFDs:
            line = runout1.readline()
            if line.startswith('server listening at '):
                port = line.rstrip().split(' ')[-1]
            else:
                failed = 1
        else:
            failed = 1
        if failed:
            print 'could not start mpdcheck server'
            print 'here is the output:'
            if line:
                print line,
            done = 0
            fds = [runout1,runerr1]
            while not done:
                try:
                    (readyFDs,unused1,unused2) = select(fds,[],[],1)
                except Exception, data:
                    print 'select 4 error: %s ; %s' % ( data.__class__, data)
                    exit(-1)
                if runout in readyFDs:
                    line = runout.readline()
                    if line:
                        print line,
                    else:
                        fds.remove(runout)
                elif runerr in readyFDs:
                    line = runerr.readline()
                    if line:
                        print line,
                    else:
                        fds.remove(runerr)
                else:
                    done = 1
        if failed:
            try:
                kill(runpid1,SIGKILL)
            except:
                pass
            exit(-1)
        cmd2 = "ssh %s -x -n %s%smpdcheck.py -c %s %s" % (host,fullDirName,path.sep,fqhn1,port)
        if verbose:
            print 'starting client: %s' % (cmd2)
        runner2 = Popen3(cmd2,1,0)
        runout2 = runner2.fromchild
        runerr2 = runner2.childerr
        runin2  = runner2.tochild
        runpid2 = runner2.pid
        try:
            (readyFDs,unused1,unused2) = select([runout2],[],[],9)
        except Exception, data:
            print 'select 3 error: %s ; %s' % ( data.__class__, data)
            exit(-1)
        if len(readyFDs) == 0:
            print '** timed out waiting for client on %s to produce output' % (host)
        line = ''
        failed = 0
        port = 0
        if runout2 in readyFDs:
            line = runout2.readline()
            if not line.startswith('client successfully recvd'):
                failed = 1
        else:
            failed = 1
        if failed:
            print 'client on %s failed to access the server' % (host)
            print 'here is the output:'
            if line:
                print line,
            done = 0
            fds = [runout2,runerr2]
            while not done:
                try:
                    (readyFDs,unused1,unused2) = select(fds,[],[],1)
                except Exception, data:
                    print 'select 4 error: %s ; %s' % ( data.__class__, data)
                    exit(-1)
                if runout2 in readyFDs:
                    line = runout2.readline()
                    if line:
                        print line,
                    else:
                        fds.remove(runout2)
                elif runerr2 in readyFDs:
                    line = runerr2.readline()
                    if line:
                        print line,
                    else:
                        fds.remove(runerr2)
                else:
                    done = 1
        try:
            kill(runpid2,SIGKILL)
        except:
            pass
        if failed:
            try:
                kill(runpid1,SIGKILL)
            except:
                pass
            exit(-1)



#!/usr/bin/env python
#
#   (C) 2001 by Argonne National Laboratory.
#       See COPYRIGHT in top-level directory.
#

"""
usage:
mpiexec [-h or -help or --help]    # get this message
mpiexec -file filename             # (or -f) filename contains XML job description
mpiexec [global args] [local args] executable [args]
   where global args may be
      -l                           # line labels by MPI rank
      -bnr                         # MPICH1 compatibility mode
      -machinefile                 # file mapping procs to machines
      -s <spec>                    # direct stdin to "all" or 1,2 or 2-4,6 
      -1                           # override default of trying 1st proc locally
      -ifhn                        # network interface to use locally
      -tv                          # run procs under totalview (must be installed)
      -tvsu                        # totalview startup only
      -gdb                         # run procs under gdb
      -m                           # merge output lines (default with gdb)
      -a                           # means assign this alias to the job
      -ecfn                        # output_xml_exit_codes_filename
      -recvtimeout <integer_val>   # timeout for recvs to fail (e.g. from mpd daemon)
      -g<local arg name>           # global version of local arg (below)
    and local args may be
      -n <n> or -np <n>            # number of processes to start
      -wdir <dirname>              # working directory to start in
      -umask <umask>               # umask for remote process
      -path <dirname>              # place to look for executables
      -host <hostname>             # host to start on
      -soft <spec>                 # modifier of -n value
      -arch <arch>                 # arch type to start on (not implemented)
      -envall                      # pass all env vars in current environment
      -envnone                     # pass no env vars
      -envlist <list of env var names> # pass current values of these vars
      -env <name> <value>          # pass this value of this env var
mpiexec [global args] [local args] executable args : [local args] executable...
mpiexec -gdba jobid                # gdb-attach to existing jobid
mpiexec -configfile filename       # filename contains cmd line segs as lines
  (See User Guide for more details)

Examples:
   mpiexec -l -n 10 cpi 100
   mpiexec -genv QPL_LICENSE 4705 -n 3 a.out

   mpiexec -n 1 -host foo master : -n 4 -host mysmp slave
"""
from time import ctime
__author__ = "Ralph Butler and Rusty Lusk"
__date__ = ctime()
__version__ = "$Revision: 1.90 $"
__credits__ = ""

import signal
if hasattr(signal,'SIGTTIN'):
    signal.signal(signal.SIGTTIN,signal.SIG_IGN)    # asap

import sys, os, socket, re

from  urllib import quote
from  time   import time
from  urllib import unquote
from  mpdlib import mpd_set_my_id, mpd_get_my_username, mpd_version, mpd_print, \
                    mpd_uncaught_except_tb, mpd_handle_signal, mpd_which, \
                    MPDListenSock, MPDStreamHandler, MPDConClientSock, MPDParmDB

try:
    import pwd
    pwd_module_available = 1
except:
    pwd_module_available = 0

global parmdb, nextRange, appnum, recvTimeout
global numDoneWithIO, myExitStatus, sigOccurred, outXmlDoc, outECs


def mpiexec():
    global parmdb, nextRange, appnum, recvTimeout
    global numDoneWithIO, myExitStatus, sigOccurred, outXmlDoc, outECs

    import sys  # for sys.excepthook on next line
    sys.excepthook = mpd_uncaught_except_tb

    myExitStatus = 0
    if len(sys.argv) < 2  or  sys.argv[1] == '-h'  \
    or  sys.argv[1] == '-help'  or  sys.argv[1] == '--help':
	usage()
    myHost = socket.gethostname()
    mpd_set_my_id(myid='mpiexec_%s' % (myHost) )
    try:
        hostinfo = socket.gethostbyname_ex(myHost)
    except:
        print 'mpiexec failed: gethostbyname_ex failed for %s' % (myHost)
        sys.exit(-1)
    myIP = hostinfo[2][0]

    parmdb = MPDParmDB(orderedSources=['cmdline','xml','env','rcfile','thispgm'])
    parmsToOverride = {
                        'MPD_USE_ROOT_MPD'            :  0,
                        'MPD_SECRETWORD'              :  '',
                        'MPIEXEC_SHOW_LINE_LABELS'    :  0,
                        'MPIEXEC_LINE_LABEL_FMT'      :  '%r',
                        'MPIEXEC_JOB_ALIAS'           :  '',
                        'MPIEXEC_USIZE'               :  0,
                        'MPIEXEC_GDB'                 :  0,
                        'MPIEXEC_IFHN'                :  '',  # use one from mpd as default
                        'MPIEXEC_MERGE_OUTPUT'        :  0,
                        'MPIEXEC_STDIN_DEST'          :  '0',
                        'MPIEXEC_MACHINEFILE'         :  '',
                        'MPIEXEC_BNR'                 :  0,
                        'MPIEXEC_TOTALVIEW'           :  0,
                        'MPIEXEC_TVSU'                :  0,
                        'MPIEXEC_EXITCODES_FILENAME'  :  '',
                        'MPIEXEC_TRY_1ST_LOCALLY'     :  1,
                        'MPIEXEC_TIMEOUT'             :  0,
                        'MPIEXEC_HOST_LIST'           :  [],
                        'MPIEXEC_HOST_CHECK'          :  0,
                        'MPIEXEC_RECV_TIMEOUT'        :  20,
                      }
    for (k,v) in parmsToOverride.items():
        parmdb[('thispgm',k)] = v
    parmdb[('thispgm','mship')] = ''
    parmdb[('thispgm','rship')] = ''
    parmdb[('thispgm','userpgm')] = ''
    parmdb[('thispgm','nprocs')] = 0
    parmdb[('thispgm','ecfn_format')] = ''
    parmdb[('thispgm','-gdba')] = ''
    parmdb[('thispgm','singinitpid')] = 0
    parmdb[('thispgm','singinitport')] = 0
    parmdb[('thispgm','ignore_rcfile')] = 0
    parmdb[('thispgm','ignore_environ')] = 0
    parmdb[('thispgm','inXmlFilename')] = ''
    parmdb[('thispgm','print_parmdb_all')] = 0
    parmdb[('thispgm','print_parmdb_def')] = 0

    appnum = 0
    nextRange = 0
    localArgSets = { 0 : [] }

    if sys.argv[1] == '-gdba':
	if len(sys.argv) != 3:
            print '-gdba arg must appear only with jobid'
	    usage()
        parmdb[('cmdline','-gdba')] = sys.argv[2]
        parmdb[('cmdline','MPIEXEC_GDB')] = 1
        parmdb[('cmdline','MPIEXEC_MERGE_OUTPUT')] = 1       # implied
        parmdb[('cmdline','MPIEXEC_SHOW_LINE_LABELS')] = 1   # implied
        parmdb[('cmdline','MPIEXEC_STDIN_DEST')]   = 'all'   # implied
    elif sys.argv[1] == '-file'  or  sys.argv[1] == '-f':
	if len(sys.argv) != 3:
            print '-file (-f) arg must appear alone'
	    usage()
        parmdb[('cmdline','inXmlFilename')] = sys.argv[2]
    elif sys.argv[1] == '-pmi_args':
        parmdb[('cmdline','singinitport')] = sys.argv[2]
        # ignoring interface name (where app is listening) and authentication key, for now
        parmdb[('cmdline','singinitpid')]  = sys.argv[5]
        parmdb[('cmdline','userpgm')] = 'unknown_pgmname'
        parmdb[('cmdline','nprocs')] = 1
        parmdb[('cmdline','MPIEXEC_TRY_1ST_LOCALLY')] = 1
        machineFileInfo = {}
        tempargv = [sys.argv[0],'unknown_pgmname']
        collect_args(tempargv,localArgSets)
    else:
        if sys.argv[1] == '-configfile':
	    if len(sys.argv) != 3:
	        usage()
            configFile = open(sys.argv[2],'r',0)
            configLines = configFile.readlines()
            configLines = [ x.strip() + ' : '  for x in configLines if x[0] != '#' and x.strip() != '' ]
            tempargv = []
            for line in configLines:
                line = 'mpddummyarg ' + line  # gets pitched in shells that can't handle --
                (shellIn,shellOut) = \
                    os.popen4("/bin/sh -c 'for a in $*; do echo _$a; done' -- %s" % (line))
                for shellLine in shellOut:
                    if shellLine.startswith('_mpddummyarg'):
                        continue
                    tempargv.append(shellLine[1:].strip())    # 1: strips off the leading _
	    tempargv = [sys.argv[0]] + tempargv[0:-1]   # strip off the last : I added
            collect_args(tempargv,localArgSets)
        else:
            collect_args(sys.argv,localArgSets)
        machineFileInfo = read_machinefile(parmdb['MPIEXEC_MACHINEFILE'])

    # set some default values for mpd; others added as discovered below
    msgToMPD = { 'cmd'            : 'mpdrun',
                 'conhost'        : myHost,
                 'spawned'        : 0,
                 'nstarted'       : 0,
                 'hosts'          : {},
                 'execs'          : {},
                 'users'          : {},
                 'cwds'           : {},
                 'umasks'         : {},
                 'paths'          : {},
                 'args'           : {},
                 'limits'         : {},
                 'envvars'        : {},
                 'ifhns'          : {},
               }

    if parmdb['inXmlFilename']:
        get_parms_from_xml_file(msgToMPD)  # fills in some more values of msgToMPD
    else:
        parmdb.get_parms_from_env(parmsToOverride)
        parmdb.get_parms_from_rcfile(parmsToOverride)

    # mostly old mpdrun below here
    numDoneWithIO = 0
    outXmlDoc = ''
    outECs = ''
    outECFile = None
    sigOccurred = 0

    recvTimeout = int(parmdb['MPIEXEC_RECV_TIMEOUT'])  # may be changed below

    listenSock = MPDListenSock('',0,name='socket_to_listen_for_man')
    listenPort = listenSock.getsockname()[1]
    if (hasattr(os,'getuid')  and  os.getuid() == 0)  or  parmdb['MPD_USE_ROOT_MPD']:
        fullDirName = os.path.abspath(os.path.split(sys.argv[0])[0])  # normalize
        mpdroot = os.path.join(fullDirName,'mpdroot')
        conSock = MPDConClientSock(mpdroot=mpdroot,secretword=parmdb['MPD_SECRETWORD'])
    else:
        conSock = MPDConClientSock(secretword=parmdb['MPD_SECRETWORD'])

    if parmdb['MPIEXEC_HOST_CHECK']:    # if this was requested in the xml file
        msgToSend = { 'cmd' : 'verify_hosts_in_ring',
                      'host_list' : parmdb['MPIEXEC_HOST_LIST'] }
        conSock.send_dict_msg(msgToSend)
        msg = conSock.recv_dict_msg(timeout=recvTimeout)
        if not msg:
            mpd_print(1,'no msg recvd from mpd for verify_hosts_in_ring')
            sys.exit(-1)
        elif msg['cmd'] != 'verify_hosts_in_ring_response':
            mpd_print(1,'unexpected msg from mpd :%s:' % (msg) )
            sys.exit(-1)
        if msg['host_list']:
            print 'These hosts are not in the mpd ring:'
            for host in  msg['host_list']:
                if host[0].isdigit():
                    print '    %s' % (host),
                    try:
                        print ' (%s)' % (socket.gethostbyaddr(host)[0])
                    except:
                        print ''
                else:
                    print '    %s' % (host)
            sys.exit(-1)

    msgToSend = { 'cmd' : 'get_mpdrun_values' }
    conSock.send_dict_msg(msgToSend)
    msg = conSock.recv_dict_msg(timeout=recvTimeout)
    if not msg:
        mpd_print(1, 'no msg recvd from mpd during version check')
        sys.exit(-1)
    elif msg['cmd'] != 'response_get_mpdrun_values':
        mpd_print(1,'unexpected msg from mpd :%s:' % (msg) )
        sys.exit(-1)
    if msg['mpd_version'] != mpd_version():
        mpd_print(1,'mpd version %s does not match mpiexec version %s' % \
                  (msg['mpd_version'],mpd_version()) )
        sys.exit(-1)

    # if using/testing the INET CONSOLE
    if os.environ.has_key('MPD_CON_INET_HOST_PORT'):
        try:
            myIfhn = socket.gethostbyname_ex(myHost)[2][0]
        except:
            print 'mpiexec failed: gethostbyname_ex failed for %s' % (myHost)
            sys.exit(-1)
        parmdb[('thispgm','MPIEXEC_IFHN')] = myIfhn
    elif not parmdb['MPIEXEC_IFHN']:    # if user did not specify one, use mpd's
        parmdb[('thispgm','MPIEXEC_IFHN')] = msg['mpd_ifhn']    # not really thispgm here

    if parmdb['-gdba']:
        get_vals_for_attach(parmdb,conSock,msgToMPD)
    elif not parmdb['inXmlFilename']:
        parmdb[('cmdline','nprocs')] = 0  # for incr later
        for k in localArgSets.keys():
	    handle_local_argset(localArgSets[k],machineFileInfo,msgToMPD)

    if parmdb['MPIEXEC_MERGE_OUTPUT']  and  not parmdb['MPIEXEC_SHOW_LINE_LABELS']:
        parmdb[('thispgm','MPIEXEC_SHOW_LINE_LABELS')] = 1   # causes line labels also

    if parmdb['print_parmdb_all']:
        parmdb.printall()
    if parmdb['print_parmdb_def']:
        parmdb.printdef()

    if parmdb['mship']:
        mshipSock = MPDListenSock('',0,name='socket_for_mship')
        mshipPort = mshipSock.getsockname()[1]
        mshipPid = os.fork()
        if mshipPid == 0:
            conSock.close()
            os.environ['MPDCP_AM_MSHIP'] = '1'
            os.environ['MPDCP_MSHIP_PORT'] = str(mshipPort)
            os.environ['MPDCP_MSHIP_FD'] = str(mshipSock.fileno())
            os.environ['MPDCP_MSHIP_NPROCS'] = str(parmdb['nprocs'])
            try:
                os.execvpe(parmdb['mship'],[parmdb['MPIEXEC_MSHIP']],os.environ)
            except Exception, errmsg:
                mpd_print(1,'execvpe failed for copgm %s; errmsg=:%s:' % \
                          (parmdb['MPIEXEC_MSHIP'],errmsg))
                sys.exit(-1)
            os._exit(0);  # do NOT do cleanup
        mshipSock.close()
    else:
        mshipPid = 0

    # make sure to do this after nprocs has its value
    linesPerRank = {}  # keep this a dict instead of a list
    for i in range(parmdb['nprocs']):
        linesPerRank[i] = []
    # make sure to do this after nprocs has its value
    if recvTimeout == 20:  # still the default
        recvTimeoutMultiplier = 0.1
        if os.environ.has_key('MPD_RECVTIMEOUT_MULTIPLIER'):
            try:
                recvTimeoutMultiplier = int(os.environ ['MPD_RECVTIMEOUT_MULTIPLIER'])
            except ValueError:
                try:
                    recvTimeoutMultiplier = float(os.environ ['MPD_RECVTIMEOUT_MULTIPLIER'])
                except ValueError:
                    print 'Invalid MPD_RECVTIMEOUT_MULTIPLIER. Value must be a number.'
                    sys.exit(-1)
        recvTimeout = int(parmdb['nprocs']) * recvTimeoutMultiplier

    if parmdb['MPIEXEC_EXITCODES_FILENAME']:
        if parmdb['ecfn_format'] == 'xml':
            try:
                import xml.dom.minidom
            except:
                print 'you requested to save the exit codes in an xml file, but'
                print '  I was unable to import the xml.dom.minidom module'
                sys.exit(-1)
            outXmlDoc = xml.dom.minidom.Document()
            outECs = outXmlDoc.createElement('exit-codes')
            outXmlDoc.appendChild(outECs)
        else:
            outECs = 'exit-codes\n'

    msgToMPD['nprocs'] = parmdb['nprocs']
    msgToMPD['limits'][(0,parmdb['nprocs']-1)]  = {}
    msgToMPD['conport'] = listenPort
    msgToMPD['conip'] = myIP
    msgToMPD['conifhn'] = parmdb['MPIEXEC_IFHN']
    if parmdb['MPIEXEC_JOB_ALIAS']:
        msgToMPD['jobalias'] = parmdb['MPIEXEC_JOB_ALIAS']
    else:
        msgToMPD['jobalias'] = ''
    if parmdb['MPIEXEC_TRY_1ST_LOCALLY']:
        msgToMPD['try_1st_locally'] = 1
    if parmdb['rship']:
        msgToMPD['rship'] = parmdb['rship']
        msgToMPD['mship_host'] = socket.gethostname()
        msgToMPD['mship_port'] = mshipPort
    if parmdb['MPIEXEC_BNR']:
        msgToMPD['doing_bnr'] = 1
    if parmdb['MPIEXEC_STDIN_DEST'] == 'all':
        stdinDest = '0-%d' % (parmdb['nprocs']-1)
    else:
        stdinDest = parmdb['MPIEXEC_STDIN_DEST']
    if parmdb['MPIEXEC_SHOW_LINE_LABELS']:
        msgToMPD['line_labels'] = parmdb['MPIEXEC_LINE_LABEL_FMT']
    else:
        msgToMPD['line_labels'] = ''
    msgToMPD['stdin_dest'] = stdinDest
    msgToMPD['gdb'] = parmdb['MPIEXEC_GDB']
    msgToMPD['gdba'] = parmdb['-gdba']
    msgToMPD['totalview'] = parmdb['MPIEXEC_TOTALVIEW']
    msgToMPD['singinitpid'] = parmdb['singinitpid']
    msgToMPD['singinitport'] = parmdb['singinitport']
    msgToMPD['host_spec_pool'] = parmdb['MPIEXEC_HOST_LIST']

    # set sig handlers up right before we send mpdrun msg to mpd
    if hasattr(signal,'SIGINT'):
        signal.signal(signal.SIGINT, sig_handler)
    if hasattr(signal,'SIGTSTP'):
        signal.signal(signal.SIGTSTP,sig_handler)
    if hasattr(signal,'SIGCONT'):
        signal.signal(signal.SIGCONT,sig_handler)
    if hasattr(signal,'SIGALRM'):
        signal.signal(signal.SIGALRM,sig_handler)

    conSock.send_dict_msg(msgToMPD)
    msg = conSock.recv_dict_msg(timeout=recvTimeout)
    if not msg:
        mpd_print(1, 'no msg recvd from mpd when expecting ack of request')
        sys.exit(-1)
    elif msg['cmd'] == 'mpdrun_ack':
        currRingSize = msg['ringsize']
        currRingNCPUs = msg['ring_ncpus']
    else:
        if msg['cmd'] == 'already_have_a_console':
            print 'mpd already has a console (e.g. for long ringtest); try later'
            sys.exit(-1)
        elif msg['cmd'] == 'job_failed':
            if  msg['reason'] == 'some_procs_not_started':
                print 'mpiexec: unable to start all procs; may have invalid machine names'
                print '    remaining specified hosts:'
                for host in msg['remaining_hosts'].values():
                    if host != '_any_':
                        try:
                            print '        %s (%s)' % (host,socket.gethostbyaddr(host)[0])
                        except:
                            print '        %s' % (host)
            elif  msg['reason'] == 'invalid_username':
                print 'mpiexec: invalid username %s at host %s' % \
                      (msg['username'],msg['host'])
            else:
                print 'mpiexec: job failed; reason=:%s:' % (msg['reason'])
            sys.exit(-1)
        else:
            mpd_print(1, 'unexpected message from mpd: %s' % (msg) )
            sys.exit(-1)
    conSock.close()
    jobTimeout = int(parmdb['MPIEXEC_TIMEOUT'])
    if jobTimeout:
        if hasattr(signal,'alarm'):
            signal.alarm(jobTimeout)
        else:
            def timeout_function():
                mpd_print(1,'job ending due to env var MPIEXEC_TIMEOUT=%d' % jobTimeout)
                thread.interrupt_main()
            try:
                import thread, threading
                timer = threading.Timer(jobTimeout,timeout_function)
                timer.start()
            except:
                print 'unable to establish timeout for MPIEXEC_TIMEOUT'

    streamHandler = MPDStreamHandler()

    (manSock,addr) = listenSock.accept()
    if not manSock:
        mpd_print(1, 'mpiexec: failed to obtain sock from manager')
        sys.exit(-1)
    streamHandler.set_handler(manSock,handle_man_input,args=(streamHandler,))
    if hasattr(os,'fork'):
        streamHandler.set_handler(sys.stdin,handle_stdin_input,
                                  args=(parmdb,streamHandler,manSock))
    else:  # not using select on fd's when using subprocess module (probably M$)
        import threading
        def read_fd_with_func(fd,func):
            line = 'x'
            while line:
                line = func(fd)
        stdin_thread = threading.Thread(target=read_fd_with_func,
                                        args=(sys.stdin.fileno(),handle_stdin_input))
    # first, do handshaking with man
    msg = manSock.recv_dict_msg()
    if (not msg  or  not msg.has_key('cmd') or msg['cmd'] != 'man_checking_in'):
        mpd_print(1, 'mpiexec: from man, invalid msg=:%s:' % (msg) )
        sys.exit(-1)
    msgToSend = { 'cmd' : 'ringsize', 'ring_ncpus' : currRingNCPUs,
                  'ringsize' : currRingSize }
    manSock.send_dict_msg(msgToSend)
    msg = manSock.recv_dict_msg()
    if (not msg  or  not msg.has_key('cmd')):
        mpd_print(1, 'mpiexec: from man, invalid msg=:%s:' % (msg) )
        sys.exit(-1)
    if (msg['cmd'] == 'job_started'):
        jobid = msg['jobid']
        if outECs:
            if parmdb['ecfn_format'] == 'xml':
                outECs.setAttribute('jobid',jobid.strip())
            else:
                outECs += 'jobid=%s\n' % (jobid.strip())
        # print 'mpiexec: job %s started' % (jobid)
        if parmdb['MPIEXEC_TVSU']:
            import mtv
            mtv.allocate_proctable(parmdb['nprocs'])
            # extract procinfo (rank,hostname,exec,pid) tuples from msg
            for i in range(parmdb['nprocs']):
                tvhost = msg['procinfo'][i][0]
                tvpgm  = msg['procinfo'][i][1]
                tvpid  = msg['procinfo'][i][2]
                # print "%d %s %s %d" % (i,host,pgm,pid)
                mtv.append_proctable_entry(tvhost,tvpgm,tvpid)
            mtv.complete_spawn()
            msgToSend = { 'cmd' : 'tv_ready' }
            manSock.send_dict_msg(msgToSend)
        elif parmdb['MPIEXEC_TOTALVIEW']:
            tvname = 'totalview'
            if os.environ.has_key('TOTALVIEW'):
                tvname = os.environ['TOTALVIEW']
            if not mpd_which(((tvname.strip()).split()[0])):
                print 'cannot find "%s" in your $PATH:' % (tvname)
                print '    ', os.environ['PATH']
                sys.exit(-1)
            import mtv
            tv_cmd = 'dattach python ' + `os.getpid()` + '; dgo; dassign MPIR_being_debugged 1'
            os.system(tvname + ' -e "%s" &' % (tv_cmd) )
            mtv.wait_for_debugger()
            mtv.allocate_proctable(parmdb['nprocs'])
            # extract procinfo (rank,hostname,exec,pid) tuples from msg
            for i in range(parmdb['nprocs']):
                tvhost = msg['procinfo'][i][0]
                tvpgm  = msg['procinfo'][i][1]
                tvpid  = msg['procinfo'][i][2]
                # print "%d %s %s %d" % (i,host,pgm,pid)
                mtv.append_proctable_entry(tvhost,tvpgm,tvpid)
            mtv.complete_spawn()
            msgToSend = { 'cmd' : 'tv_ready' }
            manSock.send_dict_msg(msgToSend)
    else:
        mpd_print(1, 'mpiexec: from man, unknown msg=:%s:' % (msg) )
        sys.exit(-1)

    (manCliStdoutSock,addr) = listenSock.accept()
    streamHandler.set_handler(manCliStdoutSock,
                              handle_cli_stdout_input,
                              args=(parmdb,streamHandler,linesPerRank,))
    (manCliStderrSock,addr) = listenSock.accept()
    streamHandler.set_handler(manCliStderrSock,
                              handle_cli_stderr_input,
                              args=(streamHandler,))

    # Main Loop
    timeDelayForPrints = 2  # seconds
    timeForPrint = time() + timeDelayForPrints   # to get started
    numDoneWithIO = 0
    while numDoneWithIO < 3:    # man, client stdout, and client stderr
        if sigOccurred:
            handle_sig_occurred(manSock)
        rv = streamHandler.handle_active_streams(timeout=1.0)
        if rv[0] < 0:  # will handle some sigs at top of next loop
            pass       # may have to handle some err conditions here
        if parmdb['MPIEXEC_MERGE_OUTPUT']:
            if timeForPrint < time():
                print_ready_merged_lines(1,parmdb,linesPerRank)
                timeForPrint = time() + timeDelayForPrints
            else:
                print_ready_merged_lines(parmdb['nprocs'],parmdb,linesPerRank)

    if parmdb['MPIEXEC_MERGE_OUTPUT']:
        print_ready_merged_lines(1,parmdb,linesPerRank)
    if mshipPid:
        (donePid,status) = os.wait()    # os.waitpid(mshipPid,0)
    if parmdb['MPIEXEC_EXITCODES_FILENAME']:
        outECFile = open(parmdb['MPIEXEC_EXITCODES_FILENAME'],'w')
        if parmdb['ecfn_format'] == 'xml':
            print >>outECFile, outXmlDoc.toprettyxml(indent='   ')
        else:
            print >>outECFile, outECs,
        outECFile.close()
    return myExitStatus


def collect_args(args,localArgSets):
    validGlobalArgs = { '-l' : 0, '-usize' : 1, '-gdb' : 0, '-bnr' : 0,
                        '-tv' : 0, '-tvsu' : 0,
                        '-ifhn' : 1, '-machinefile' : 1, '-s' : 1, '-1' : 0,
                        '-a' : 1, '-m' : 0, '-ecfn' : 1, '-recvtimeout' : 1,
                        '-gn' : 1, '-gnp' : 1, '-ghost' : 1, '-gpath' : 1, '-gwdir' : 1,
			'-gsoft' : 1, '-garch' : 1, '-gexec' : 1, '-gumask' : 1,
			'-genvall' : 0, '-genv' : 2, '-genvnone' : 0,
			'-genvlist' : 1 }
    currumask = os.umask(0) ; os.umask(currumask)  # grab it and set it back
    parmdb[('cmdline','-gn')]          = 1
    parmdb[('cmdline','-ghost')]       = '_any_'
    if os.environ.has_key('PATH'):
        parmdb[('cmdline','-gpath')]   = os.environ['PATH']
    else:
        parmdb[('cmdline','-gpath')]   =  ''
    parmdb[('cmdline','-gwdir')]       = os.path.abspath(os.getcwd())
    parmdb[('cmdline','-gumask')]      = str(currumask)
    parmdb[('cmdline','-gsoft')]       = 0
    parmdb[('cmdline','-garch')]       = ''
    parmdb[('cmdline','-gexec')]       = ''
    parmdb[('cmdline','-genv')]        = {}
    parmdb[('cmdline','-genvlist')]    = []
    parmdb[('cmdline','-genvnone')]    = 0
    argidx = 1
    while argidx < len(args)  and  args[argidx] in validGlobalArgs.keys():
        garg = args[argidx]
        if len(args) <= (argidx+validGlobalArgs[garg]):
            print "missing sub-arg to %s" % (garg)
            usage()
        if garg == '-genv':
            parmdb['-genv'][args[argidx+1]] = args[argidx+2]
            argidx += 3
        elif garg == '-gn'  or  garg == '-gnp':
            if args[argidx+1].isdigit():
                parmdb[('cmdline','-gn')] = int(args[argidx+1])
            else:
                print 'argument to %s must be numeric' % (garg)
                usage()
            argidx += 2
        elif garg == '-ghost':
            try:
                parmdb[('cmdline',garg)] = socket.gethostbyname_ex(args[argidx+1])[2][0]
            except:
                print 'unable to do find info for host %s' % (args[argidx+1])
                sys.exit(-1)
            argidx += 2
        elif garg == '-gpath':
            parmdb[('cmdline','-gpath')] = args[argidx+1]
            argidx += 2
        elif garg == '-gwdir':
            parmdb[('cmdline','-gwdir')] = args[argidx+1]
            argidx += 2
        elif garg == '-gumask':
            parmdb[('cmdline','-gumask')] = args[argidx+1]
            argidx += 2
        elif garg == '-gsoft':
            parmdb[('cmdline','-gsoft')] = args[argidx+1]
            argidx += 2
        elif garg == '-garch':
            parmdb[('cmdline','-garch')] = args[argidx+1]
            argidx += 2
            print '** -garch is accepted but not used'
        elif garg == '-gexec':
            parmdb[('cmdline','-gexec')] = args[argidx+1]
            argidx += 2
        elif garg == '-genv':
            parmdb[('cmdline','-genv')] = args[argidx+1]
            argidx += 2
        elif garg == '-genvlist':
            parmdb[('cmdline','-genvlist')] = args[argidx+1].split(',')
            argidx += 2
        elif garg == '-genvnone':
            parmdb[('cmdline','-genvnone')] = args[argidx+1]
            argidx += 1
        elif garg == '-l':
            parmdb[('cmdline','MPIEXEC_SHOW_LINE_LABELS')] = 1
            argidx += 1
        elif garg == '-a':
            parmdb[('cmdline','MPIEXEC_JOB_ALIAS')] = args[argidx+1]
            argidx += 2
        elif garg == '-usize':
            if args[argidx+1].isdigit():
                parmdb[('cmdline','MPIEXEC_USIZE')] = int(args[argidx+1])
            else:
                print 'argument to %s must be numeric' % (garg)
                usage()
            argidx += 2
        elif garg == '-recvtimeout':
            if args[argidx+1].isdigit():
                parmdb[('cmdline','MPIEXEC_RECV_TIMEOUT')] = int(args[argidx+1])
            else:
                print 'argument to %s must be numeric' % (garg)
                usage()
            argidx += 2
        elif garg == '-gdb':
            parmdb[('cmdline','MPIEXEC_GDB')] = 1
            argidx += 1
            parmdb[('cmdline','MPIEXEC_MERGE_OUTPUT')] = 1       # implied
            parmdb[('cmdline','MPIEXEC_SHOW_LINE_LABELS')] = 1   # implied
            parmdb[('cmdline','MPIEXEC_STDIN_DEST')]   = 'all'   # implied
        elif garg == '-ifhn':
            parmdb[('cmdline','MPIEXEC_IFHN')] = args[argidx+1]
            argidx += 2
            try:
                hostinfo = socket.gethostbyname_ex(parmdb['MPIEXEC_IFHN'])
            except:
                print 'mpiexec: gethostbyname_ex failed for ifhn %s' % (parmdb['MPIEXEC_IFHN'])
                sys.exit(-1)
        elif garg == '-m':
            parmdb[('cmdline','MPIEXEC_MERGE_OUTPUT')] = 1
            argidx += 1
        elif garg == '-s':
            parmdb[('cmdline','MPIEXEC_STDIN_DEST')] = args[argidx+1]
            argidx += 2
        elif garg == '-machinefile':
            parmdb[('cmdline','MPIEXEC_MACHINEFILE')] = args[argidx+1]
            argidx += 2
        elif garg == '-bnr':
            parmdb[('cmdline','MPIEXEC_BNR')] = 1
            argidx += 1
        elif garg == '-tv':
            parmdb[('cmdline','MPIEXEC_TOTALVIEW')] = 1
            argidx += 1
        elif garg == '-tvsu':
            parmdb[('cmdline','MPIEXEC_TOTALVIEW')] = 1
            parmdb[('cmdline','MPIEXEC_TVSU')] = 1
            argidx += 1
        elif garg == '-ecfn':
            parmdb[('cmdline','MPIEXEC_EXITCODES_FILENAME')] = args[argidx+1]
            argidx += 2
        elif garg == '-1':
            parmdb[('cmdline','MPIEXEC_TRY_1ST_LOCALLY')] = 0  # reverses meaning
            argidx += 1
    if len(args) <= argidx:
        print "mpiexec: missing arguments after global args"
        usage()
    if args[argidx] == ':':
        argidx += 1
    localArgsKey = 0
    # collect local arg sets but do not validate them until handled below
    while argidx < len(args):
        if args[argidx] == ':':
            localArgsKey += 1
            localArgSets[localArgsKey] = []
        else:
            localArgSets[localArgsKey].append(args[argidx])
        argidx += 1

def handle_local_argset(argset,machineFileInfo,msgToMPD):
    global parmdb, nextRange, appnum, recvTimeout
    validLocalArgs  = { '-n' : 1, '-np' : 1, '-host' : 1, '-path' : 1, '-wdir' : 1,
                        '-soft' : 1, '-arch' : 1, '-umask' : 1,
			'-envall' : 0, '-env' : 2, '-envnone' : 0, '-envlist' : 1 }
    host   = parmdb['-ghost']
    wdir   = parmdb['-gwdir']
    wumask = parmdb['-gumask']
    wpath  = parmdb['-gpath']
    nProcs = parmdb['-gn']
    usize  = parmdb['MPIEXEC_USIZE']
    gexec  = parmdb['-gexec']
    softness =  parmdb['-gsoft']
    if parmdb['-genvnone']:
        envall = 0
    else:
        envall = 1
    localEnvlist = []
    localEnv  = {}
    
    argidx = 0
    while argidx < len(argset):
        if argset[argidx] not in validLocalArgs:
            if argset[argidx][0] == '-':
                print 'invalid "local" arg: %s' % argset[argidx]
                usage()
            break                       # since now at executable
        if parmdb['MPIEXEC_MACHINEFILE']:
            if argset[argidx] == '-host'  or  argset[argidx] == ['-ghost']:
                print '-host (or -ghost) and -machinefile are incompatible'
                sys.exit(-1)
        if argset[argidx] == '-n' or argset[argidx] == '-np':
            if len(argset) < (argidx+2):
                print '** missing arg to -n'
                usage()
            nProcs = argset[argidx+1]
            if not nProcs.isdigit():
                print '** non-numeric arg to -n: %s' % nProcs
                usage()
            nProcs = int(nProcs)
            argidx += 2
        elif argset[argidx] == '-host':
            if len(argset) < (argidx+2):
                print '** missing arg to -host'
                usage()
            try:
                host = socket.gethostbyname_ex(argset[argidx+1])[2][0]
            except:
                print 'unable to do find info for host %s' % (argset[argidx+1])
                sys.exit(-1)
            argidx += 2
        elif argset[argidx] == '-path':
            if len(argset) < (argidx+2):
                print '** missing arg to -path'
                usage()
            wpath = argset[argidx+1]
            argidx += 2
        elif argset[argidx] == '-wdir':
            if len(argset) < (argidx+2):
                print '** missing arg to -wdir'
                usage()
            wdir = argset[argidx+1]
            argidx += 2
        elif argset[argidx] == '-umask':
            if len(argset) < (argidx+2):
                print '** missing arg to -umask'
                usage()
            wumask = argset[argidx+1]
            argidx += 2
        elif argset[argidx] == '-soft':
            if len(argset) < (argidx+2):
                print '** missing arg to -soft'
                usage()
            softness = argset[argidx+1]
            argidx += 2
        elif argset[argidx] == '-arch':
            if len(argset) < (argidx+2):
                print '** missing arg to -arch'
                usage()
            print '** -arch is accepted but not used'
            argidx += 2
        elif argset[argidx] == '-envall':
            envall = 1
            argidx += 1
        elif argset[argidx] == '-envnone':
            envall = 0
            argidx += 1
        elif argset[argidx] == '-envlist':
            localEnvlist = argset[argidx+1].split(',')
            argidx += 2
        elif argset[argidx] == '-env':
            if len(argset) < (argidx+3):
                print '** missing arg to -env'
                usage()
            var = argset[argidx+1]
            val = argset[argidx+2]
            localEnv[var] = val
            argidx += 3
        else:
            print 'unknown "local" option: %s' % argset[argidx]
            usage()

    if softness:
        nProcs = adjust_nprocs(nProcs,softness)

    cmdAndArgs = []
    if argidx < len(argset):
        while argidx < len(argset):
            cmdAndArgs.append(argset[argidx])
            argidx += 1
    else:
        if gexec:
            cmdAndArgs = [gexec]
    if not cmdAndArgs:
        print 'no cmd specified'
        usage()

    argsetLoRange = nextRange
    argsetHiRange = nextRange + nProcs - 1
    loRange = argsetLoRange
    hiRange = argsetHiRange

    defaultHostForArgset = host
    while loRange <= argsetHiRange:
        host = defaultHostForArgset
        if machineFileInfo:
            if len(machineFileInfo) <= hiRange:
                print 'too few entries in machinefile'
                sys.exit(-1)
            host = machineFileInfo[loRange]['host']
            ifhn = machineFileInfo[loRange]['ifhn']
            if ifhn:
                msgToMPD['ifhns'][loRange] = ifhn
            for i in range(loRange+1,hiRange+1):
                if machineFileInfo[i]['host'] != host  or  machineFileInfo[i]['ifhn'] != ifhn:
                    hiRange = i - 1
                    break

        asRange = (loRange,hiRange)  # this argset range as a tuple

        msgToMPD['users'][asRange]  = mpd_get_my_username()
        msgToMPD['execs'][asRange]  = cmdAndArgs[0]
        msgToMPD['paths'][asRange]  = wpath
        msgToMPD['cwds'][asRange]   = wdir
        msgToMPD['umasks'][asRange] = wumask
        msgToMPD['args'][asRange]   = cmdAndArgs[1:]
        if host.startswith('_any_'):
            msgToMPD['hosts'][(loRange,hiRange)] = host
        else:
            try:
                msgToMPD['hosts'][asRange] = socket.gethostbyname_ex(host)[2][0]
            except:
                print 'unable to do find info for host %s' % (host)
                sys.exit(-1)

        envToSend = {}
        if envall:
            for envvar in os.environ.keys():
                envToSend[envvar] = os.environ[envvar]
        for envvar in parmdb['-genvlist']:
            if not os.environ.has_key(envvar):
                print '%s in envlist does not exist in your env' % (envvar)
                sys.exit(-1)
            envToSend[envvar] = os.environ[envvar]
        for envvar in localEnvlist:
            if not os.environ.has_key(envvar):
                print '%s in envlist does not exist in your env' % (envvar)
                sys.exit(-1)
            envToSend[envvar] = os.environ[envvar]
        for envvar in parmdb['-genv'].keys():
            envToSend[envvar] = parmdb['-genv'][envvar]
        for envvar in localEnv.keys():
            envToSend[envvar] = localEnv[envvar]
        if usize:
            envToSend['MPI_UNIVERSE_SIZE'] = str(usize)
        envToSend['MPI_APPNUM'] = str(appnum)
        msgToMPD['envvars'][(loRange,hiRange)] = envToSend

        loRange = hiRange + 1
        hiRange = argsetHiRange  # again

    appnum += 1
    nextRange += nProcs
    parmdb[('cmdline','nprocs')] = parmdb['nprocs'] + nProcs
    
# Adjust nProcs (called maxprocs in the Standard) according to soft:
# Our interpretation is that we need the largest number <= nProcs that is
# consistent with the list of possible values described by soft.  I.e.
# if the user says
#
#   mpiexec -n 10 -soft 5 a.out
#
# we adjust the 10 down to 5.  This may not be what was intended in the Standard,
# but it seems to be what it says.

def adjust_nprocs(nProcs,softness):
    biglist = []
    list1 = softness.split(',')
    for triple in list1:                # triple is a or a:b or a:b:c
        thingy = triple.split(':')     
        if len(thingy) == 1:
            a = int(thingy[0])
            if a <= nProcs and a >= 0:
                biglist.append(a)
        elif len(thingy) == 2:
            a = int(thingy[0])
            b = int(thingy[1])
            for i in range(a,b+1):
                if i <= nProcs and i >= 0:
                    biglist.append(i)
        elif len(thingy) == 3:
            a = int(thingy[0])
            b = int(thingy[1])
            c = int(thingy[2])
            for i in range(a,b+1,c):
                if i <= nProcs and i >= 0:
                    biglist.append(i)                
        else:
            print 'invalid subargument to -soft: %s' % (softness)
            print 'should be a or a:b or a:b:c'
            usage()

        if len(biglist) == 0:
            print '-soft argument %s allows no valid number of processes' % (softness)
            usage()
        else:
            return max(biglist)


def read_machinefile(machineFilename):
    if not machineFilename:
        return None
    try:
        machineFile = open(machineFilename,'r')
    except:
        print '** unable to open machinefile'
        sys.exit(-1)
    procID = 0
    machineFileInfo = {}
    for line in machineFile:
        line = line.strip()
        if not line  or  line[0] == '#':
            continue
        splitLine = re.split(r'\s+',line)
        host = splitLine[0]
        if ':' in host:
            (host,nprocs) = host.split(':',1)
            nprocs = int(nprocs)
        else:
            nprocs = 1
        kvps = {'ifhn' : ''}
        for kv in splitLine[1:]:
            (k,v) = kv.split('=',1)
            if k == 'ifhn':  # interface hostname
                kvps[k] = v
            else:  # may be other kv pairs later
                print 'unrecognized key in machinefile:', k
                sys.exit(-1)
        for i in range(procID,procID+nprocs):
            machineFileInfo[i] = { 'host' : host, 'nprocs' : nprocs }
            machineFileInfo[i].update(kvps)
        procID += nprocs
    return machineFileInfo

def handle_man_input(sock,streamHandler):
    global numDoneWithIO, myExitStatus
    global outXmlDoc, outECs
    msg = sock.recv_dict_msg()
    if not msg:
        streamHandler.del_handler(sock)
        numDoneWithIO += 1
    elif not msg.has_key('cmd'):
        mpd_print(1,'mpiexec: from man, invalid msg=:%s:' % (msg) )
        sys.exit(-1)
    elif msg['cmd'] == 'startup_status':
        if msg['rc'] != 0:
            # print 'rank %d (%s) in job %s failed to find executable %s' % \
                  # ( msg['rank'], msg['src'], msg['jobid'], msg['exec'] )
            host = msg['src'].split('_')[0]
            reason = unquote(msg['reason'])
            print 'problem with execution of %s  on  %s:  %s ' % \
                  (msg['exec'],host,reason)
            # don't stop ; keep going until all top-level mans finish
    elif msg['cmd'] == 'job_aborted_early':
        print 'rank %d in job %s caused collective abort of all ranks' % \
              ( msg['rank'], msg['jobid'] )
        status = msg['exit_status']
        if hasattr(os,'WIFSIGNALED')  and  os.WIFSIGNALED(status):
            if status > myExitStatus:
                myExitStatus = status
            killed_status = status & 0x007f  # AND off core flag
            print '  exit status of rank %d: killed by signal %d ' % \
                  (msg['rank'],killed_status)
        elif hasattr(os,'WEXITSTATUS'):
            exit_status = os.WEXITSTATUS(status)
            if exit_status > myExitStatus:
                myExitStatus = exit_status
            print '  exit status of rank %d: return code %d ' % \
                  (msg['rank'],exit_status)
        else:
            myExitStatus = 0
    elif msg['cmd'] == 'job_aborted':
        print 'job aborted; reason = %s' % (msg['reason'])
    elif msg['cmd'] == 'client_exit_status':
        if outECs:
            if parmdb['ecfn_format'] == 'xml':
                outXmlProc = outXmlDoc.createElement('exit-code')
                outECs.appendChild(outXmlProc)
                outXmlProc.setAttribute('rank',str(msg['cli_rank']))
                outXmlProc.setAttribute('status',str(msg['cli_status']))
                outXmlProc.setAttribute('pid',str(msg['cli_pid']))
                outXmlProc.setAttribute('host',msg['cli_host'])  # cli_ifhn is also avail
            else:
                outECs += 'rank=%d status=%d pid=%d host=%s\n' % \
                          (msg['cli_rank'],msg['cli_status'],msg['cli_pid'],msg['cli_host'])

        # print "exit info: rank=%d  host=%s  pid=%d  status=%d" % \
              # (msg['cli_rank'],msg['cli_host'],
               # msg['cli_pid'],msg['cli_status'])
        status = msg['cli_status']
        if hasattr(os,'WIFSIGNALED')  and  os.WIFSIGNALED(status):
            if status > myExitStatus:
                myExitStatus = status
            killed_status = status & 0x007f  # AND off core flag
            # print 'exit status of rank %d: killed by signal %d ' % \
            #        (msg['cli_rank'],killed_status)
        elif hasattr(os,'WEXITSTATUS'):
            exit_status = os.WEXITSTATUS(status)
            if exit_status > myExitStatus:
                myExitStatus = exit_status
            # print 'exit status of rank %d: return code %d ' % \
            #       (msg['cli_rank'],exit_status)
        else:
            myExitStatus = 0
    else:
        print 'unrecognized msg from manager :%s:' % msg

def handle_cli_stdout_input(sock,parmdb,streamHandler,linesPerRank):
    global numDoneWithIO
    if parmdb['MPIEXEC_MERGE_OUTPUT']:
        line = sock.recv_one_line()
        if not line:
            streamHandler.del_handler(sock)
            numDoneWithIO += 1
        else:
            if parmdb['MPIEXEC_GDB']:
                line = line.replace('(gdb)\n','(gdb) ')
            try:
                (rank,rest) = line.split(':',1)
                rank = int(rank)
                linesPerRank[rank].append(rest)
            except:
                print line
            print_ready_merged_lines(parmdb['nprocs'],parmdb,linesPerRank)
    else:
        msg = sock.recv(1024)
        if not msg:
            streamHandler.del_handler(sock)
            numDoneWithIO += 1
        else:
            sys.stdout.write(msg)
            sys.stdout.flush()

def handle_cli_stderr_input(sock,streamHandler):
    global numDoneWithIO
    msg = sock.recv(1024)
    if not msg:
        streamHandler.del_handler(sock)
        numDoneWithIO += 1
    else:
        sys.stderr.write(msg)
        sys.stderr.flush()

# NOTE: stdin is supposed to be slow, low-volume.  We read it all here (as it
# appears on the fd) and send it immediately to the receivers.  If the user 
# redirects a "large" file (perhaps as small as 5k) into us, we will send it
# all out right away.  This can cause things to hang on the remote (recvr) side.
# We do not wait to read here until the recvrs read because there may be several
# recvrs and they may read at different speeds/times.
def handle_stdin_input(stdin_stream,parmdb,streamHandler,manSock):
    line  = ''
    try:
        line = stdin_stream.readline()
    except IOError, errinfo:
        sys.stdin.flush()  # probably does nothing
        # print "I/O err on stdin:", errinfo
        mpd_print(1,'stdin problem; if pgm is run in background, redirect from /dev/null')
        mpd_print(1,'    e.g.: mpiexec -n 4 a.out < /dev/null &')
    else:
        gdbFlag = parmdb['MPIEXEC_GDB']
        if line:    # not EOF
            msgToSend = { 'cmd' : 'stdin_from_user', 'line' : line } # default
            if gdbFlag and line.startswith('z'):
                line = line.rstrip()
                if len(line) < 3:    # just a 'z'
                    line += ' 0-%d' % (parmdb['nprocs']-1)
                s1 = line[2:].rstrip().split(',')
                for s in s1:
                    s2 = s.split('-')
                    for i in s2:
                        if not i.isdigit():
                            print 'invalid arg to z :%s:' % i
                            continue
                msgToSend = { 'cmd' : 'stdin_dest', 'stdin_procs' : line[2:] }
                sys.stdout.softspace = 0
                print '%s:  (gdb) ' % (line[2:]),
            elif gdbFlag and line.startswith('q'):
                msgToSend = { 'cmd' : 'stdin_dest',
                              'stdin_procs' : '0-%d' % (parmdb['nprocs']-1) }
                if manSock:
                    manSock.send_dict_msg(msgToSend)
                msgToSend = { 'cmd' : 'stdin_from_user','line' : 'q\n' }
            elif gdbFlag and line.startswith('^'):
                msgToSend = { 'cmd' : 'stdin_dest',
                              'stdin_procs' : '0-%d' % (parmdb['nprocs']-1) }
                if manSock:
                    manSock.send_dict_msg(msgToSend)
                msgToSend = { 'cmd' : 'signal', 'signo' : 'SIGINT' }
            if manSock:
                manSock.send_dict_msg(msgToSend)
        else:
            streamHandler.del_handler(sys.stdin)
            sys.stdin.close()
            if manSock:
                msgToSend = { 'cmd' : 'stdin_from_user', 'eof' : '' }
                manSock.send_dict_msg(msgToSend)
    return line

def handle_sig_occurred(manSock):
    global sigOccurred
    if sigOccurred == signal.SIGINT:
        if manSock:
            msgToSend = { 'cmd' : 'signal', 'signo' : 'SIGINT' }
            manSock.send_dict_msg(msgToSend)
            manSock.close()
        sys.exit(-1)
    elif sigOccurred == signal.SIGALRM:
        if manSock:
            msgToSend = { 'cmd' : 'signal', 'signo' : 'SIGKILL' }
            manSock.send_dict_msg(msgToSend)
            manSock.close()
        mpd_print(1,'job ending due to env var MPIEXEC_TIMEOUT=%s' % \
                  os.environ['MPIEXEC_TIMEOUT'])
        sys.exit(-1)
    elif sigOccurred == signal.SIGTSTP:
        sigOccurred = 0  # do this before kill below
        if manSock:
            msgToSend = { 'cmd' : 'signal', 'signo' : 'SIGTSTP' }
            manSock.send_dict_msg(msgToSend)
        signal.signal(signal.SIGTSTP,signal.SIG_DFL)      # stop myself
        os.kill(os.getpid(),signal.SIGTSTP)
        signal.signal(signal.SIGTSTP,sig_handler)  # restore this handler
    elif sigOccurred == signal.SIGCONT:
        sigOccurred = 0  # do it before handling
        if manSock:
            msgToSend = { 'cmd' : 'signal', 'signo' : 'SIGCONT' }
            manSock.send_dict_msg(msgToSend)

def sig_handler(signum,frame):
    global sigOccurred
    sigOccurred = signum
    mpd_handle_signal(signum,frame)

def format_sorted_ranks(ranks):
    all = []
    one = []
    prevRank = -999
    for i in range(len(ranks)):
        if i != 0  and  ranks[i] != (prevRank+1):
            all.append(one)
            one = []
        one.append(ranks[i])
        if i == (len(ranks)-1):
            all.append(one)
        prevRank = ranks[i]
    pline = ''
    for i in range(len(all)):
        if len(all[i]) > 1:
            pline += '%d-%d' % (all[i][0],all[i][-1])
        else:
            pline += '%d' % (all[i][0])
        if i != (len(all)-1):
            pline += ','
    return pline

def print_ready_merged_lines(minRanks,parmdb,linesPerRank):
    printFlag = 1  # default to get started
    while printFlag:
        printFlag = 0
        for r1 in range(parmdb['nprocs']):
            if not linesPerRank[r1]:
                continue
            sortedRanks = []
            lineToPrint = linesPerRank[r1][0]
            for r2 in range(parmdb['nprocs']):
                if linesPerRank[r2] and linesPerRank[r2][0] == lineToPrint: # myself also
                    sortedRanks.append(r2)
            if len(sortedRanks) >= minRanks:
                fsr = format_sorted_ranks(sortedRanks)
                sys.stdout.softspace = 0
                print '%s: %s' % (fsr,lineToPrint),
                for r2 in sortedRanks:
                    linesPerRank[r2] = linesPerRank[r2][1:]
                printFlag = 1
    sys.stdout.flush()

def get_parms_from_xml_file(msgToMPD):
    global parmdb
    try:
        import xml.dom.minidom
    except:
        print 'you requested to parse an xml file, but'
        print '  I was unable to import the xml.dom.minidom module'
        sys.exit(-1)
    known_rlimit_types = ['core','cpu','fsize','data','stack','rss',
                          'nproc','nofile','ofile','memlock','as','vmem']
    try:
        inXmlFilename = parmdb['inXmlFilename'] 
        parmsXMLFile = open(inXmlFilename,'r')
    except:
        print 'could not open job xml specification file %s' % (inXmlFilename)
        sys.exit(-1)
    fileContents = parmsXMLFile.read()
    try:
        parsedXML = xml.dom.minidom.parseString(fileContents)
    except:
        print "mpiexec failed parsing xml file (perhaps from mpiexec); here is the content:"
        print fileContents
        sys.exit(-1)
    if parsedXML.documentElement.tagName != 'create-process-group':
        print 'expecting create-process-group; got unrecognized doctype: %s' % \
              (parsedXML.documentElement.tagName)
        sys.exit(-1)
    cpg = parsedXML.getElementsByTagName('create-process-group')[0]
    if cpg.hasAttribute('totalprocs'):
        parmdb[('xml','nprocs')] = int(cpg.getAttribute('totalprocs'))
    else:
        print '** totalprocs not specified in %s' % inXmlFilename
        sys.exit(-1)
    if cpg.hasAttribute('try_1st_locally'):
        parmdb[('xml','MPIEXEC_TRY_1ST_LOCALLY')] = int(cpg.getAttribute('try_1st_locally'))
    if cpg.hasAttribute('output')  and  cpg.getAttribute('output') == 'label':
        parmdb[('xml','MPIEXEC_SHOW_LINE_LABELS')] = 1
    if cpg.hasAttribute('pgid'):    # our jobalias
        parmdb[('xml','MPIEXEC_JOB_ALIAS')] = cpg.getAttribute('pgid')
    if cpg.hasAttribute('stdin_dest'):
        parmdb[('xml','MPIEXEC_STDIN_DEST')] = cpg.getAttribute('stdin_dest')
    if cpg.hasAttribute('doing_bnr'):
        parmdb[('xml','MPIEXEC_BNR')] = int(cpg.getAttribute('doing_bnr'))
    if cpg.hasAttribute('ifhn'):
        parmdb[('xml','MPIEXEC_IFHN')] = cpg.getAttribute('ifhn')
    if cpg.hasAttribute('exit_codes_filename'):
        parmdb[('xml','MPIEXEC_EXITCODES_FILENAME')] = cpg.getAttribute('exit_codes_filename')
        parmdb[('xml','ecfn_format')] = 'xml'
    if cpg.hasAttribute('gdb'):
        gdbFlag = int(cpg.getAttribute('gdb'))
        if gdbFlag:
            parmdb[('xml','MPIEXEC_GDB')]     = 1
            parmdb[('xml','MPIEXEC_MERGE_OUTPUT')] = 1       # implied
            parmdb[('xml','MPIEXEC_SHOW_LINE_LABELS')] = 1   # implied
            parmdb[('xml','MPIEXEC_STDIN_DEST')]   = 'all'   # implied
    if cpg.hasAttribute('use_root_pm'):
        parmdb[('xml','MPD_USE_ROOT_MPD')] = int(cpg.getAttribute('use_root_pm'))
    if cpg.hasAttribute('tv'):
        parmdb[('xml','MPIEXEC_TOTALVIEW')] = int(cpg.getAttribute('tv'))
    hostSpec = cpg.getElementsByTagName('host-spec')
    if hostSpec:
        hostList = []
        for node in hostSpec[0].childNodes:
            node = node.data.strip()
            hostnames = re.findall(r'\S+',node)
            for hostname in hostnames:
                if hostname:    # some may be the empty string
                    try:
                        ipaddr = socket.gethostbyname_ex(hostname)[2][0]
                    except:
                        print 'unable to determine IP info for host %s' % (hostname)
                        sys.exit(-1)
                    hostList.append(ipaddr)
        parmdb[('xml','MPIEXEC_HOST_LIST')] = hostList
    if hostSpec and hostSpec[0].hasAttribute('check'):
        hostSpecMode = hostSpec[0].getAttribute('check')
        if hostSpecMode == 'yes':
            parmdb[('xml','MPIEXEC_HOST_CHECK')] = 1
    covered = [0] * parmdb['nprocs'] 
    procSpec = cpg.getElementsByTagName('process-spec')
    if not procSpec:
        print 'No process-spec specified'
        usage()
    for p in procSpec:
        if p.hasAttribute('range'):
            therange = p.getAttribute('range')
            splitRange = therange.split('-')
            if len(splitRange) == 1:
                loRange = int(splitRange[0])
                hiRange = loRange
            else:
                (loRange,hiRange) = (int(splitRange[0]),int(splitRange[1]))
        else:
            (loRange,hiRange) = (0,parmdb['nprocs']-1)
        for i in xrange(loRange,hiRange+1):
            nprocs = parmdb['nprocs']
            if i >= nprocs:
                print '*** exiting; rank %d is greater than nprocs' % (nprocs)
                sys.exit(-1)
            if covered[i]:
                print '*** exiting; rank %d is doubly used in proc specs' % (nprocs)
                sys.exit(-1)
            covered[i] = 1
        if p.hasAttribute('exec'):
            msgToMPD['execs'][(loRange,hiRange)] = p.getAttribute('exec')
        else:
            print '*** exiting; range %d-%d has no exec' % (loRange,hiRange)
            sys.exit(-1)
        if p.hasAttribute('user'):
            username = p.getAttribute('user')
            if pwd_module_available:
                try:
                    pwent = pwd.getpwnam(username)
                except:
                    print username, 'is an invalid username'
                    sys.exit(-1)
            if username == mpd_get_my_username()  \
            or (hasattr(os,'getuid') and os.getuid() == 0):
                msgToMPD['users'][(loRange,hiRange)] = p.getAttribute('user')
            else:
                print username, 'username does not match yours and you are not root'
                sys.exit(-1)
        else:
            msgToMPD['users'][(loRange,hiRange)] = mpd_get_my_username()
        if p.hasAttribute('cwd'):
            msgToMPD['cwds'][(loRange,hiRange)] = p.getAttribute('cwd')
        else:
            msgToMPD['cwds'][(loRange,hiRange)] = os.path.abspath(os.getcwd())
        if p.hasAttribute('umask'):
            msgToMPD['umasks'][(loRange,hiRange)] = p.getAttribute('umask')
        else:
            currumask = os.umask(0) ; os.umask(currumask)
            msgToMPD['umasks'][(loRange,hiRange)] = str(currumask)
        if p.hasAttribute('path'):
            msgToMPD['paths'][(loRange,hiRange)] = p.getAttribute('path')
        else:
            msgToMPD['paths'][(loRange,hiRange)] = os.environ['PATH']
        if p.hasAttribute('host'):
            host = p.getAttribute('host')
            if host.startswith('_any_'):
                msgToMPD['hosts'][(loRange,hiRange)] = host
            else:
                try:
                    msgToMPD['hosts'][(loRange,hiRange)] = socket.gethostbyname_ex(host)[2][0]
                except:
                    print 'unable to do find info for host %s' % (host)
                    sys.exit(-1)
        else:
            if hostSpec  and  hostList:
                msgToMPD['hosts'][(loRange,hiRange)] = '_any_from_pool_'
            else:
                msgToMPD['hosts'][(loRange,hiRange)] = '_any_'
        argDict = {}
        argList = p.getElementsByTagName('arg')
        for argElem in argList:
            argDict[int(argElem.getAttribute('idx'))] = argElem.getAttribute('value')
        argVals = [0] * len(argList)
        for i in argDict.keys():
            argVals[i-1] = unquote(argDict[i])
        msgToMPD['args'][(loRange,hiRange)] = argVals
        limitDict = {}
        limitList = p.getElementsByTagName('limit')
        for limitElem in limitList:
            typ = limitElem.getAttribute('type')
            if typ in known_rlimit_types:
                limitDict[typ] = limitElem.getAttribute('value')
            else:
                print 'mpiexec: invalid type in limit: %s' % (typ)
                sys.exit(-1)
        msgToMPD['limits'][(loRange,hiRange)] = limitDict
        envVals = {}
        envVarList = p.getElementsByTagName('env')
        for envVarElem in envVarList:
            envkey = envVarElem.getAttribute('name')
            envval = unquote(envVarElem.getAttribute('value'))
            envVals[envkey] = envval
        msgToMPD['envvars'][(loRange,hiRange)] = envVals
    for i in range(len(covered)):
        if not covered[i]:
            print '*** exiting; %d procs are requested, but proc %d is not described' % \
                  (parmdb['nprocs'],i)
            sys.exit(-1)
        
def get_vals_for_attach(parmdb,conSock,msgToMPD):
    global recvTimeout
    sjobid = parmdb['-gdba'].split('@')    # jobnum and originating host
    msgToSend = { 'cmd' : 'mpdlistjobs' }
    conSock.send_dict_msg(msgToSend)
    msg = conSock.recv_dict_msg(timeout=recvTimeout)
    if not msg:
        mpd_print(1,'no msg recvd from mpd before timeout')
        sys.exit(-1)
    if msg['cmd'] != 'local_mpdid':     # get full id of local mpd for filters later
        mpd_print(1,'did not recv local_mpdid msg from local mpd; recvd: %s' % msg)
        sys.exit(-1)
    else:
        if len(sjobid) == 1:
            sjobid.append(msg['id'])
    got_info = 0
    while 1:
        msg = conSock.recv_dict_msg()
        if not msg.has_key('cmd'):
            mpd_print(1,'invalid message from mpd :%s:' % (msg))
            sys.exit(-1)
        if msg['cmd'] == 'mpdlistjobs_info':
            got_info = 1
            smjobid = msg['jobid'].split('  ')  # jobnum, mpdid, and alias (if present)
            if sjobid[0] == smjobid[0]  and  sjobid[1] == smjobid[1]:  # jobnum and mpdid
                rank = int(msg['rank'])
                msgToMPD['users'][(rank,rank)]   = msg['username']
                msgToMPD['hosts'][(rank,rank)]   = msg['ifhn']
                msgToMPD['execs'][(rank,rank)]   = msg['pgm']
                msgToMPD['cwds'][(rank,rank)]    = os.path.abspath(os.getcwd())
                msgToMPD['paths'][(rank,rank)]   = os.environ['PATH']
                msgToMPD['args'][(rank,rank)]    = [msg['clipid']]
                msgToMPD['envvars'][(rank,rank)] = {}
                msgToMPD['limits'][(rank,rank)]  = {}
                currumask = os.umask(0) ; os.umask(currumask)  # grab it and set it back
                msgToMPD['umasks'][(rank,rank)]  = str(currumask)
        elif  msg['cmd'] == 'mpdlistjobs_trailer':
            if not got_info:
                print 'no info on this jobid; probably invalid'
                sys.exit(-1)
            break
        else:
            print 'invaild msg from mpd :%s:' % (msg)
            sys.exit(-1)
    parmdb[('thispgm','nprocs')] = len(msgToMPD['execs'].keys())  # all dicts are same len
    

def usage():
    print __doc__
    sys.exit(-1)


if __name__ == '__main__':
    try:
        mpiexec()
    except SystemExit, errExitStatus:  # bounced to here by sys.exit inside mpiexec()
        myExitStatus = errExitStatus
    sys.exit(myExitStatus)
